# mypy: disable-error-code="attr-defined"
import asyncio

from bluesky.protocols import Triggerable
from frappy.errors import ImpossibleError
from ophyd_async.core import SignalR, SignalX

from secop_ophyd.SECoPDevices import (
    SECoPCMDDevice,
    SECoPMoveableDevice,
    SECoPNodeDevice,
)


async def test_struct_inp_cmd(nested_struct_sim, nested_node_no_re: SECoPNodeDevice):
    test_cmd: SECoPCMDDevice = nested_node_no_re.ophy_struct.test_cmd_CMD

    input_dict = {"name": "test_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    run_obj: SignalX = test_cmd.commandx

    stat = run_obj.trigger()

    await stat

    reading_res = await res.read()
    print(reading_res)
    assert isinstance(reading_res.get(res.name)["value"], int)


def test_triggerable(nested_struct_sim, nested_node_no_re: SECoPNodeDevice):
    test_cmd: SECoPCMDDevice = nested_node_no_re.ophy_struct.test_cmd_CMD

    assert isinstance(test_cmd, Triggerable)


async def test_secop_error_on_cmd(
    nested_struct_sim, nested_node_no_re: SECoPNodeDevice
):
    test_cmd: SECoPCMDDevice = nested_node_no_re.ophy_struct.test_cmd_CMD

    error_triggered = False
    # Triggers SECoP Error
    input_dict = {"name": "bad_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    run_obj: SignalX = test_cmd.commandx

    try:
        stat = run_obj.trigger()
        await stat

    except ImpossibleError:
        error_triggered = True

    assert error_triggered is True

    reading_res = await res.read()
    assert reading_res.get(res.name)["value"] is None


async def test_secop_triggering_cmd_dev(
    nested_struct_sim, nested_node_no_re: SECoPNodeDevice
):
    test_cmd: SECoPCMDDevice = nested_node_no_re.ophy_struct.test_cmd_CMD

    input_dict = {"name": "test_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    stat = test_cmd.trigger()

    await stat

    reading_res = await res.read()
    assert isinstance(reading_res.get(res.name)["value"], int)


async def test_stop_cmd(cryo_sim, cryo_node_no_re: SECoPNodeDevice):
    cryo: SECoPMoveableDevice = cryo_node_no_re.cryo

    await cryo.window.set(5)

    await cryo.tolerance.set(1)

    await cryo.ramp.set(20)

    stat = cryo.set(15)

    await asyncio.sleep(3)

    # essentially a NOOP (stops are only passed through to SECoP on success=False)
    await cryo.stop(success=True)

    assert cryo._stopped is False, "Move should not be stopped when success=True"
    assert (
        cryo._success is True
    ), "Move should be marked as successful when success=True"
    assert (
        not stat.done
    ), "Move should still be in progress after stop with success=True"

    # move is still going on
    await cryo.stop(success=False)
    assert cryo._stopped is True, "Move should be stopped when success=False"
    assert (
        cryo._success is False
    ), "Move should be marked as unsuccessful when success=False"

    await stat



import logging
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import Dict, Optional

# Default configuration
DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_LOG_DIR = ".secop-ophyd"
DEFAULT_LOG_FILENAME = "secop-ophyd.log"
DEFAULT_ROTATION_WHEN = "H"  # Rotate every hour
DEFAULT_ROTATION_INTERVAL = 1  # Every 1 hour
DEFAULT_BACKUP_COUNT = 48  # Keep logs for 48 hours

log_file_handlers: Dict[str, TimedRotatingFileHandler] = {}

console_handler: logging.StreamHandler | None = None


def setup_logging(
    name: str = "secop_ophyd",
    level: int = DEFAULT_LOG_LEVEL,
    log_dir: Optional[str] = None,
) -> logging.Logger:
    """
    Set up and configure a logger with timed rotating file handler.

    Parameters:
    ----------
    name : str
        Name of the logger
    level : int
        Logging level (default: INFO)
    log_format : str
        Format string for log messages
    log_dir : Optional[str]
        Directory to store log files (default: .secop-ophyd in current directory)
    log_file : Optional[str]
        Log file name (default: secop-ophyd.log)
    when : str
        Type of interval - can be 'S' (seconds), 'M' (minutes), 'H' (hours),
        'D' (days), 'W0'-'W6' (weekday, 0=Monday), 'midnight'
    interval : int
        How many units between rotations (default: 1 hour)
    backup_count : int
        Number of backup files to keep (default: 48 - for 48 hours of logs)
    console : bool
        Whether to also log to console (default: False)

    Returns:
    -------
    logging.Logger
        Configured logger instance
    """

    log_format: str = DEFAULT_LOG_FORMAT

    log_file: Optional[str] = None
    when: str = DEFAULT_ROTATION_WHEN
    interval: int = DEFAULT_ROTATION_INTERVAL
    backup_count: int = DEFAULT_BACKUP_COUNT

    # Create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # If handlers already exist, don't add more
    if logger.handlers:
        return logger

    # Create formatter
    formatter = logging.Formatter(log_format)

    # Set up log directory
    if log_dir is None:
        log_dir = DEFAULT_LOG_DIR

    log_path = Path(log_dir)
    log_path.mkdir(parents=True, exist_ok=True)

    # Set up log file path
    if log_file is None:
        log_file = DEFAULT_LOG_FILENAME

    log_file_path = log_path / log_file
    # Use resolved string path as key to ensure consistency
    handler_key = str(log_file_path.resolve())

    if handler_key not in log_file_handlers:
        # Create timed rotating file handler
        file_handler = TimedRotatingFileHandler(
            log_file_path,
            when=when,
            interval=interval,
            backupCount=backup_count,
            encoding="utf-8",
            utc=True,  # Use UTC time for consistency
        )
        file_handler.setFormatter(formatter)
        log_file_handlers[handler_key] = file_handler
        is_new_handler = True

    else:
        # Use existing handler (don't overwrite formatter)
        file_handler = log_file_handlers[handler_key]
        is_new_handler = False

    # Add console handler if requested
    # global console_handler

    # if not console_handler:
    #     console_handler = logging.StreamHandler(sys.stdout)
    #     console_handler.setFormatter(formatter)
    #     logger.addHandler(console_handler)
    # else:
    #     logger.addHandler(console_handler)

    logger.addHandler(file_handler)

    # Log initialization message
    if is_new_handler:
        logger.info(f"Logger '{name}' initialized with new log file: {log_file_path}")
    else:
        logger.info(f"Logger '{name}' added to shared log file: {log_file_path}")

    return logger


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger by name. If the logger doesn't have handlers, it will be created
    with default settings.

    Parameters:
    ----------
    name : str
        Name of the logger

    Returns:
    -------
    logging.Logger
        Logger instance
    """
    logger = logging.getLogger(name)

    # If the logger doesn't have handlers, set it up with defaults
    if not logger.handlers:
        return setup_logging(name)

    return logger



import asyncio
import warnings
from functools import wraps
from typing import Any, Callable

from bluesky.protocols import DataKey, Reading
from frappy.client import CacheItem
from frappy.datatypes import (
    ArrayOf,
    BLOBType,
    BoolType,
    DataType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from ophyd_async.core import (
    Callback,
    SignalBackend,
    SignalDatatypeT,
    StandardReadableFormat,
    StrictEnum,
)

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.util import Path, SECoPDataKey, SECoPdtype, SECoPReading, deep_get

atomic_dtypes = (
    StringType,
    ScaledInteger,
    IntRange,
    FloatRange,
    BoolType,
    BLOBType,
    ArrayOf,
)


# max depth for datatypes supported by tiled/databroker
MAX_DEPTH = 1


class AttributeType(StrictEnum):
    PARAMETER = "parameter"
    PROPERTY = "property"


class LocalBackend(SignalBackend):
    """Class for the 'argument' and 'result' Signal backends of a SECoP_CMD_Device.
    These Signals act as a local cache for storing the command argument and result.

    """

    def __init__(
        self, path: Path, secop_dtype_obj: DataType, sig_datainfo: dict
    ) -> None:
        """Initialize SECoP_CMD_IO_Backend

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param SECoPdtype_obj: detailed SECoP datatype object for bidirectional
        conversion between JSON to and numpy arrays
        :type SECoPdtype_obj: DataType
        :param sig_datainfo: SECoP datainfo string of the value represented
        by the signal
        :type sig_datainfo: dict
        """
        self.SECoP_type_info: SECoPdtype = SECoPdtype(secop_dtype_obj)

        self.reading: SECoPReading = SECoPReading(
            secop_dt=self.SECoP_type_info, entry=None
        )

        # module:acessible Path for reading/writing (module,accessible)
        self.path: Path = path

        # Root datainfo or memberinfo for nested datatypes
        # TODO check if this is really needed
        self.datainfo: dict = sig_datainfo

        self.callback: Callback | None = None

        self.SECoPdtype_obj: DataType = secop_dtype_obj

        self.describe_dict: dict

        self.source_name = self.path._module_name + ":" + self.path._accessible_name

        self.describe_dict = {}

        self.describe_dict["source"] = self.source("", True)

        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        for property_name, prop_val in self.datainfo.items():
            if property_name == "type":
                property_name = "SECoP_dtype"
            self.describe_dict[property_name] = prop_val

        super().__init__(datatype=self.SECoP_type_info.np_datatype)

    def source(self, name: str, read: bool) -> str:
        return self.source_name

    async def connect(self, timeout: float):
        pass

    async def put(self, value: Any | None):
        self.reading.set_reading(self.SECoP_type_info.val2secop(value))

        if self.callback is not None:
            self.callback(self.reading.get_reading())

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""
        return describedict_to_datakey(self.describe_dict)

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        return self.reading.get_reading()

    async def get_value(self) -> SignalDatatypeT:
        return self.reading.get_value()

    async def get_setpoint(self) -> SignalDatatypeT:
        return await self.get_value()

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        self.callback = callback  # type: ignore[assignment]


class SECoPXBackend(SignalBackend):
    """
    Signal backend for SignalX of a SECoP_CMD_Device, that handles command execution

    """

    def __init__(
        self,
        path: Path,
        secclient: AsyncFrappyClient,
        argument: LocalBackend | None,
        result: LocalBackend | None,
    ) -> None:
        """Initializes SECoP_CMD_X_Backend

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param argument: Refence to Argument Signal
        :type argument: SECoP_CMD_IO_Backend | None
        :param result: Reference to Result Signal
        :type result: SECoP_CMD_IO_Backend | None
        """

        self._secclient: AsyncFrappyClient = secclient

        # module:acessible Path for reading/writing (module,accessible)
        self.path: Path = path

        self.callback: Callable
        self.argument: LocalBackend | None = argument
        self.result: LocalBackend | None = result

        self.source_name = self.path._module_name + ":" + self.path._accessible_name
        super().__init__(datatype=None)

    def source(self, name: str, read: bool) -> str:
        return self.source_name

    async def connect(self, timeout: float):
        pass

    async def put(self, value: Any | None):

        if self.argument is None:
            argument = None
        else:
            argument = await self.argument.get_value()

        res, qualifiers = await asyncio.wait_for(
            fut=self._secclient.exec_command(
                module=self.path._module_name,
                command=self.path._accessible_name,
                argument=argument,
            ),
            timeout=None,
        )

        # write return Value to corresponding Backend

        if self.result is None:
            return
        else:
            val = self.result.SECoP_type_info.secop2val(res)

            await self.result.put(val)

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""

        return DataKey(shape=[], dtype="string", source=self.source("", True))

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )

    async def get_value(self) -> SignalDatatypeT:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        pass

    async def get_setpoint(self) -> SignalDatatypeT:
        raise Exception(
            "Cannot read _x Signal, it has no value and is only"
            + " used to trigger Command execution"
        )


class SECoPBackend(SignalBackend[SignalDatatypeT]):
    """Unified backend for SECoP Parameters and Properties.


    This allows a single backend type to be used in signal_backend_factory,
    with deferred initialization based on annotation metadata.
    """

    format: StandardReadableFormat
    attribute_type: str | None
    _module_name: str | None
    _attribute_name: str | None  # parameter or property name
    _secclient: AsyncFrappyClient
    path_str: str
    SECoPdtype_obj: DataType
    SECoP_type_info: SECoPdtype
    describe_dict: dict

    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        path: str | None = None,
        attribute_type: str | None = None,
        secclient: AsyncFrappyClient | None = None,
    ):
        """Initialize backend (supports deferred initialization).

        Args:
            datatype: Optional datatype for the signal
            path: Optional path for immediate initialization (module:param or prop_key)
            secclient: Optional SECoP client for immediate initialization
        """
        self._module_name = None
        self._attribute_name = None

        self.attribute_type = attribute_type

        if secclient:
            self._secclient = secclient

        self.path_str = path or ""

        if path and secclient:

            if path.count(":") == 0:
                self._module_name = None
                self._attribute_name = path
            else:
                self._module_name, self._attribute_name = path.split(":", maxsplit=1)

        super().__init__(datatype)

    def init_parameter_from_introspection(
        self,
        datatype: type[SignalDatatypeT],
        path: str,
        secclient: AsyncFrappyClient,
    ):
        if self.attribute_type is not None:

            if secclient != self._secclient:
                raise RuntimeError(
                    "Backend already initialized with a different SECoP client, cannot "
                    "re-initialize"
                )

            if self.attribute_type != AttributeType.PARAMETER:
                raise RuntimeError(
                    f"Backend already initialized as {self.attribute_type}, "
                    f"cannot re-initialize as PARAMETER"
                )

        self.attribute_type = AttributeType.PARAMETER

        module_name, parameter_name = path.split(":", maxsplit=1)

        self._module_name = module_name
        self._attribute_name = parameter_name
        self._secclient = secclient

        self.datatype = datatype

        self.path_str = path

    def init_property_from_introspection(
        self, datatype: type[SignalDatatypeT], path: str, secclient: AsyncFrappyClient
    ):
        if self.attribute_type is not None:

            if secclient != self._secclient:
                raise RuntimeError(
                    "Backend already initialized with a different SECoP client, cannot "
                    "re-initialize"
                )

            if self.attribute_type != AttributeType.PROPERTY:
                raise RuntimeError(
                    f"Backend already initialized as {self.attribute_type}, cannot "
                    f"re-initialize as PROPERTY"
                )

        self.attribute_type = AttributeType.PROPERTY
        if path.count(":") == 0:
            module_name = None
            property_name = path
        else:
            module_name, property_name = path.split(":", maxsplit=1)

        self._module_name = module_name
        self._attribute_name = property_name

        self._secclient = secclient
        self.datatype = datatype

        self.path_str = path

    def source(self, name: str, read: bool) -> str:
        return self._secclient.host + ":" + self._secclient.port + ":" + self.path_str

    async def connect(self, timeout: float):
        """Connect and initialize backend (handles both parameters and properties)."""
        await self._secclient.connect()

        match self.attribute_type:
            case AttributeType.PROPERTY:
                await self._init_property()
            case AttributeType.PARAMETER:
                await self._init_parameter()

    async def _init_parameter(self):
        """Initialize as a parameter signal."""
        self._param_description: dict = self._get_param_desc()

        if not hasattr(self, "format"):
            match self._param_description.get("_signal_format", None):
                case "HINTED_SIGNAL":
                    self.format = StandardReadableFormat.HINTED_SIGNAL
                case "HINTED_UNCACHED_SIGNAL":
                    self.format = StandardReadableFormat.HINTED_UNCACHED_SIGNAL
                case "UNCACHED_SIGNAL":
                    self.format = StandardReadableFormat.UNCACHED_SIGNAL
                case _:
                    self.format = StandardReadableFormat.CONFIG_SIGNAL

        # Root datainfo or memberinfo for nested datatypes
        self.datainfo: dict = self._param_description["datainfo"]
        self.readonly = self._param_description.get("readonly")
        self.SECoPdtype_obj = self._param_description["datatype"]
        self.SECoP_type_info = SECoPdtype(self.SECoPdtype_obj)

        if self.SECoP_type_info.max_depth > MAX_DEPTH:
            warnings.warn(
                f"The datatype of parameter '{self._attribute_name}' has a maximum "
                f"depth of {self.SECoP_type_info.max_depth}. Tiled & Databroker only "
                f"support a Depth upto {MAX_DEPTH} "
                f"dtype_descr: {self.SECoP_type_info.dtype_descr}"
            )

        self.source_name = (
            self._secclient.uri
            + ":"
            + self._secclient.nodename
            + ":"
            + self._module_name
            + ":"
            + self._attribute_name
        )

        self.describe_dict = {}
        self.describe_dict["source"] = self.source_name
        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        for property_name, prop_val in self._param_description.items():
            if property_name in ("datainfo", "datatype"):
                continue
            self.describe_dict[property_name] = prop_val

        for property_name, prop_val in self.datainfo.items():
            if property_name == "type":
                property_name = "SECoP_dtype"
            if property_name == "unit":
                property_name = "units"
            self.describe_dict[property_name] = prop_val

        self.datatype = self.SECoP_type_info.np_datatype

    async def _init_property(self):
        """Initialize as a property signal."""

        if self._module_name:
            module_desc = self._secclient.modules[self._module_name]
            self._property_dict = module_desc["properties"]
        else:
            self._property_dict = self._secclient.properties

        self._prop_value = self._property_dict[self._attribute_name]
        self.SECoPdtype_obj = secop_dtype_obj_from_json(self._prop_value)
        self.SECoP_type_info = SECoPdtype(self.SECoPdtype_obj)

        if self.SECoP_type_info.max_depth > MAX_DEPTH:
            warnings.warn(
                f"The datatype of property '{self._attribute_name}' has a maximum "
                f"depth of {self.SECoP_type_info.max_depth}. Tiled & Databroker only "
                f"support a Depth upto {MAX_DEPTH} "
                f"dtype_descr: {self.SECoP_type_info.dtype_descr}"
            )

        self.describe_dict = {}
        self.describe_dict["source"] = self.path_str
        self.describe_dict.update(self.SECoP_type_info.get_datakey())

        # Properties are always readonly
        self.format = StandardReadableFormat.CONFIG_SIGNAL
        self.readonly = True
        self.datatype = self.SECoP_type_info.np_datatype

    async def put(self, value: Any | None):
        """Put a value to the parameter. Properties are readonly."""

        if self.attribute_type == AttributeType.PROPERTY:
            # Properties are readonly
            raise RuntimeError(
                f"Cannot set property '{self._attribute_name}', properties are readonly"
            )

        # convert to frappy compatible Format
        secop_val = self.SECoP_type_info.val2secop(value)

        await self._secclient.set_parameter(**self.get_param_path(), value=secop_val)

    async def get_datakey(self, source: str) -> DataKey:
        """Metadata like source, dtype, shape, precision, units"""
        if self.attribute_type == AttributeType.PROPERTY:
            # Properties have static metadata
            return describedict_to_datakey(self.describe_dict)

        if self.SECoP_type_info._is_composite or isinstance(
            self.SECoPdtype_obj, ArrayOf
        ):
            # getlast cached value
            dataset = await self._secclient.get_parameter(
                **self.get_param_path(), trycache=True
            )

            # this ensures the datakey is updated to the latest cached value
            SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)
            self.describe_dict.update(self.SECoP_type_info.get_datakey())

        return describedict_to_datakey(self.describe_dict)

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        """Get reading, handling both parameters and properties."""
        if self.attribute_type == AttributeType.PROPERTY:
            # Properties have static values
            dataset = CacheItem(
                value=self._prop_value, timestamp=self._secclient.conn_timestamp
            )
            sec_reading = SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)
            return sec_reading.get_reading()

        else:
            # Parameters are fetched from SECoP
            dataset = await self._secclient.get_parameter(
                **self.get_param_path(), trycache=True
            )
            sec_reading = SECoPReading(entry=dataset, secop_dt=self.SECoP_type_info)
            return sec_reading.get_reading()

    async def get_value(self) -> SignalDatatypeT:
        dataset: Reading = await self.get_reading()
        return dataset["value"]  # type: ignore

    async def get_setpoint(self) -> SignalDatatypeT:
        return await self.get_value()

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        if self.attribute_type == AttributeType.PROPERTY:
            # Properties are static, no callbacks
            return

        def awaitify(sync_func):
            """Wrap a synchronous callable to allow ``await``'ing it"""

            @wraps(sync_func)
            async def async_func(*args, **kwargs):
                return sync_func(*args, **kwargs)

            return async_func

        def updateItem(module, parameter, entry: CacheItem):  # noqa: N802
            data = SECoPReading(secop_dt=self.SECoP_type_info, entry=entry)
            async_callback = awaitify(callback)

            asyncio.run_coroutine_threadsafe(
                async_callback(reading=data.get_reading()),
                self._secclient.loop,
            )

        if callback is not None:
            self._secclient.register_callback(self.get_path_tuple(), updateItem)
        else:
            self._secclient.unregister_callback(self.get_path_tuple(), updateItem)

    def _get_param_desc(self) -> dict:
        return deep_get(
            self._secclient.modules,
            [self._module_name, "parameters", self._attribute_name],
        )

    def get_param_path(self):
        return {"module": self._module_name, "parameter": self._attribute_name}

    def get_path_tuple(self):
        return (self._module_name, self._attribute_name)


def secop_dtype_obj_from_json(prop_val):
    if isinstance(prop_val, str):
        return StringType()

    if isinstance(prop_val, (int, float)):
        return FloatRange()

    if isinstance(prop_val, bool):
        return BoolType()

    if isinstance(prop_val, dict):  # SECoP Structs/tuples --> numpy ndarray
        members = {}
        for key, elem in prop_val.items():
            members[key] = secop_dtype_obj_from_json(elem)

        return StructOf(**members)

    if isinstance(prop_val, list):
        # empty list, cannot infer proper type
        if not prop_val:
            return ArrayOf(FloatRange())
        # check if all elements have same Type:
        if all(isinstance(elem, type(prop_val[0])) for elem in prop_val):
            members = secop_dtype_obj_from_json(prop_val[0])
            return ArrayOf(members)
        else:
            members = []  # type: ignore
            for elem in prop_val:
                members.append(secop_dtype_obj_from_json(elem))  # type: ignore
            return TupleOf(*members)

    raise Exception(
        f"""unsupported datatype in Property:  {str(prop_val.__class__.__name__)}\n
        propval: {prop_val}"""
    )


def describedict_to_datakey(describe_dict: dict) -> SECoPDataKey:
    """Convert a DataKey to a SECoPDataKey"""
    datakey = SECoPDataKey(
        dtype=describe_dict["dtype"],
        shape=describe_dict["shape"],
        source=describe_dict["source"],
        SECOP_datainfo=describe_dict["SECOP_datainfo"],
    )

    if "units" in describe_dict:
        datakey["units"] = describe_dict["units"]

    if "dtype_str" in describe_dict:
        datakey["dtype_str"] = describe_dict["dtype_str"]

    if "dtype_descr" in describe_dict:
        datakey["dtype_descr"] = describe_dict["dtype_descr"]

    if "dtype_numpy" in describe_dict:
        datakey["dtype_numpy"] = describe_dict["dtype_numpy"]

    return datakey



# mypy: disable-error-code="attr-defined"


from ophyd_async.core import SignalR, init_devices

from secop_ophyd.SECoPDevices import SECoPMoveableDevice, SECoPNodeDevice


async def test_subset_signals_annotation(cryo_sim):

    from typing import Annotated as A

    from numpy import ndarray
    from ophyd_async.core import SignalRW
    from ophyd_async.core import StandardReadableFormat as Format
    from ophyd_async.core import StrictEnum

    from secop_ophyd.SECoPDevices import ParameterType

    class Cryostat_Mode_Enum(StrictEnum):
        """mode enum for `Cryostat`."""

        RAMP = "ramp"
        PID = "pid"
        OPENLOOP = "openloop"

    class Cryostat(SECoPMoveableDevice):
        """A simulated cc cryostat with heat-load, specific heat for the sample and a
        temperature dependent heat-link between sample and regulation."""

        # Module Parameters
        value: A[
            SignalR[float], ParameterType(), Format.HINTED_SIGNAL
        ]  # regulation temperature; Unit: (K)
        status: A[SignalR[ndarray], ParameterType()]  # current status of the module
        target: A[
            SignalRW[float], ParameterType(), Format.HINTED_SIGNAL
        ]  # target temperature; Unit: (K)

    class Cryo_7_frappy_demo(SECoPNodeDevice):
        """short description

        This is a very long description providing all the gory details about the stuff
        we are describing.
        """

        # Module Devices
        cryo: Cryostat

    async with init_devices():
        cryo = Cryo_7_frappy_demo(
            sec_node_uri="localhost:10769",
        )

    val_read = await cryo.cryo.read()
    assert val_read != {}


async def test_enum_annotation(cryo_sim):

    from typing import Annotated as A

    from numpy import ndarray
    from ophyd_async.core import SignalR, SignalRW
    from ophyd_async.core import StandardReadableFormat as Format
    from ophyd_async.core import StrictEnum

    from secop_ophyd.SECoPDevices import (
        ParameterType,
        PropertyType,
        SECoPMoveableDevice,
        SECoPNodeDevice,
    )

    class Cryostat_Mode_Enum(StrictEnum):
        """mode enum for `Cryostat`."""

        RAMP = "ramp"
        PID = "pid"
        OPENLOOP = "openloop"

    class Cryostat(SECoPMoveableDevice):
        """A simulated cc cryostat with heat-load, specific heat for the sample and a
        temperature dependent heat-link between sample and regulation."""

        # Module Properties
        group: A[SignalR[str], PropertyType()]
        description: A[SignalR[str], PropertyType()]
        implementation: A[SignalR[str], PropertyType()]
        interface_classes: A[SignalR[ndarray], PropertyType()]
        features: A[SignalR[ndarray], PropertyType()]

        # Module Parameters
        value: A[
            SignalR[float], ParameterType(), Format.HINTED_SIGNAL
        ]  # regulation temperature; Unit: (K)
        status: A[SignalR[ndarray], ParameterType()]  # current status of the module
        target: A[
            SignalRW[float], ParameterType(), Format.HINTED_SIGNAL
        ]  # target temperature; Unit: (K)
        ramp: A[
            SignalRW[float], ParameterType()
        ]  # ramping speed of the setpoint; Unit: (K/min)
        setpoint: A[
            SignalR[float], ParameterType()
        ]  # current setpoint during ramping else target; Unit: (K)
        mode: A[SignalRW[Cryostat_Mode_Enum], ParameterType()]  # mode of regulation
        maxpower: A[SignalRW[float], ParameterType()]  # Maximum heater power; Unit: (W)
        heater: A[SignalR[float], ParameterType()]  # current heater setting; Unit: (%)
        heaterpower: A[
            SignalR[float], ParameterType()
        ]  # current heater power; Unit: (W)
        pid: A[SignalRW[ndarray], ParameterType()]  # regulation coefficients
        p: A[
            SignalRW[float], ParameterType()
        ]  # regulation coefficient 'p'; Unit: (%/K)
        i: A[SignalRW[float], ParameterType()]  # regulation coefficient 'i'
        d: A[SignalRW[float], ParameterType()]  # regulation coefficient 'd'
        tolerance: A[
            SignalRW[float], ParameterType()
        ]  # temperature range for stability checking; Unit: (K)
        window: A[
            SignalRW[float], ParameterType()
        ]  # time window for stability checking; Unit: (s)
        timeout: A[
            SignalRW[float], ParameterType()
        ]  # max waiting time for stabilisation check; Unit: (s)

    class Cryo_7_frappy_demo(SECoPNodeDevice):
        """short description

        This is a very long description providing all the gory details about the stuff
        we are describing.
        """

        # Module Devices
        cryo: Cryostat

        # Node Properties
        equipment_id: A[SignalR[str], PropertyType()]
        firmware: A[SignalR[str], PropertyType()]
        description: A[SignalR[str], PropertyType()]
        _interfaces: A[SignalR[ndarray], PropertyType()]

    async with init_devices():
        cryo = Cryo_7_frappy_demo(
            sec_node_uri="localhost:10769",
        )

    await cryo.cryo.mode.set(Cryostat_Mode_Enum.RAMP)
    mode_read = await cryo.cryo.mode.get_value()

    print(mode_read)
    assert mode_read == Cryostat_Mode_Enum.RAMP



"""Code generation for annotated ophyd device classes using Jinja2 templates."""

import ast
import inspect
import io
import json
import linecache
import re
import sys
import textwrap
import tokenize
from dataclasses import dataclass, field
from enum import StrEnum
from importlib import import_module, reload
from inspect import Signature
from logging import Logger
from pathlib import Path
from types import ModuleType
from typing import get_type_hints

import autoflake
import black
from frappy.client import get_datatype
from frappy.datatypes import DataType
from jinja2 import Environment, PackageLoader, select_autoescape
from ophyd_async.core import Signal, SignalR, SignalRW, StandardReadable
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.core._utils import get_origin_class

from secop_ophyd.SECoPDevices import (
    IGNORED_PROPS,
    ParameterType,
    PropertyType,
    class_from_interface,
    secop_enum_name_to_python,
)
from secop_ophyd.SECoPSignal import secop_dtype_obj_from_json
from secop_ophyd.util import SECoPdtype


def internalize_name(name: str) -> str:
    """how to create internal names"""
    if name.startswith("_"):
        return name[1:]
    return name


@dataclass
class EnumMember:
    """Represents an enum member with name and value."""

    name: str  # Python identifier (e.g., "LOW")
    value: str  # Original SECoP string (e.g., "Low Energy")
    description: str | None = None  # Optional description


@dataclass
class EnumClass:
    """Represents an enum class definition."""

    name: str  # Enum class name (e.g., "TemperatureRegulatorModeEnum")
    members: list[EnumMember]
    description: str | None = None  # Optional enum description
    base_enum_class: str = "StrictEnum"  # "StrictEnum" or "SupersetEnum"


@dataclass
class ModuleAttribute:
    """Represents a module attribute with name, type, and optional description."""

    name: str
    type: str


@dataclass
class PropertyAttribute:
    """Represents a module property attribute with name, type"""

    name: str
    type_param: str | None = (
        None  # Optional type parameter like float for SignalRW[float]
    )
    path_annotation: str | None = (
        None  # Annotation like ParamPath(...) or PropPath(...)
    )

    type: str = "SignalR"  # Default to SignalR for properties


@dataclass
class ParameterAttribute:
    """Represents a module parameter attribute with name, type, and
    optional description."""

    name: str
    type: str
    type_param: str | None = (
        None  # Optional type parameter like float for SignalRW[float]
    )
    description: str | None = None  # Optional description from SECoP or docstrings

    path_annotation: str | None = (
        None  # Annotation like ParamPath(...) or PropPath(...)
    )
    format_annotation: str | None = None  # StandardReadableFormat.CONFIG_SIGNAL, etc.


class Method:
    """Represents a class method with signature and description.

    This class supports both old-style initialization (for backward compatibility)
    and new-style dataclass initialization.
    """

    def __init__(self, cmd_name: str, description: str, cmd_sign: Signature) -> None:
        """Initialize Method (backward compatibility constructor).

        Args:
            cmd_name: Name of the command
            description: Description of the command
            cmd_sign: Signature of the command
        """
        raw_sig_str: str = str(cmd_sign)
        raw_sig_str = raw_sig_str.replace("typing.", "")

        if "self" in raw_sig_str:
            sig_str = raw_sig_str
        else:
            sig_str = "(self, " + raw_sig_str[1:]

        self.name: str = cmd_name
        self.signature: str = sig_str
        self.description: str = description


@dataclass
class ModuleClass:
    """Represents a module class to be generated."""

    name: str
    bases: list[str]
    parameters: list[ParameterAttribute] = field(default_factory=list)
    properties: list[PropertyAttribute] = field(default_factory=list)
    methods: list[Method] = field(default_factory=list)
    description: str = ""
    enums: list[EnumClass] = field(default_factory=list)  # Enum classes for this module


@dataclass
class NodeClass:
    """Represents a node class to be generated."""

    name: str
    bases: list[str]
    properties: list[PropertyAttribute] = field(default_factory=list)
    modules: list[ModuleAttribute] = field(default_factory=list)
    description: str = ""


class GenNodeCode:
    """Generates annotated Python classes for SECoP ophyd devices.

    This class can generate Python code in two ways:
    1. By introspecting a fully instantiated SECoP ophyd device
    2. From a SECoP JSON describe message (future feature)

    The generated code uses Jinja2 templates and is formatted with Black.
    """

    ModName: str = "genNodeClass"

    def __init__(self, path: str | None = None, log=None):
        """Initialize the code generator.

        Args:
            path: Optional path to the module folder
            log: Optional logger instance
        """
        self.log: Logger | None = log
        self.module_folder_path: Path | None = None
        if path is not None:
            self.module_folder_path = Path(path)

        # Data structures for classes and imports
        self.imports: dict[str, set[str] | None] = {}
        self.module_classes: list[ModuleClass] = []
        self.node_classes: list[NodeClass] = []
        self.enum_classes: list[EnumClass] = []
        self.node_mod: ModuleType | None = None
        self.inline_comment_threshold: int = 120
        self.comment_wrap_width: int = 100

        # Required imports for generated classes
        self.add_import("typing", "Annotated as A")
        self.add_import("ophyd_async.core", "SignalR")
        self.add_import("ophyd_async.core", "SignalRW")
        self.add_import("ophyd_async.core", "SignalX")
        self.add_import("ophyd_async.core", "StandardReadableFormat as Format")
        self.add_import("ophyd_async.core", "StrictEnum")
        self.add_import("ophyd_async.core", "SupersetEnum")
        self.add_import("typing", "Any")
        self.add_import("numpy", "ndarray")
        self.add_import("secop_ophyd.SECoPDevices", "ParameterType as ParamT")
        self.add_import("secop_ophyd.SECoPDevices", "PropertyType as PropT")
        # Add necessary Device imports
        self.add_import("secop_ophyd.SECoPDevices", "SECoPDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPCommunicatorDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPReadableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPTriggerableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPWritableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPMoveableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPNodeDevice")

        # Setup Jinja2 environment
        self.jinja_env = Environment(
            loader=PackageLoader("secop_ophyd", "templates"),
            autoescape=select_autoescape(),
            trim_blocks=False,
            lstrip_blocks=False,
            keep_trailing_newline=True,
        )

        # Try to load existing generated module
        self._load_existing_module()

    def _load_existing_module(self):
        """Load existing generated module if present."""

        mod_path = self.ModName

        if self.module_folder_path is not None:
            # For absolute paths, we need to add to sys.path and import just the module
            # name
            if self.module_folder_path.is_absolute():
                str_path = str(self.module_folder_path)
                if str_path not in sys.path:
                    sys.path.insert(0, str_path)
                # Just use the module name when the folder is in sys.path
                mod_path = self.ModName
            else:
                # For relative paths, construct the module path with dots
                str_path = str(self.module_folder_path)
                rep_slash = str_path.replace("/", ".").replace("\\", ".")
                mod_path = f"{rep_slash}.{self.ModName}"

        # Remove cached module to ensure fresh import (important when module file
        # has been modified or recreated between imports)
        if mod_path in sys.modules:
            del sys.modules[mod_path]

        # Clear linecache for the module file to ensure inspect.getsource() works
        if self.module_folder_path is not None:
            module_file = self.module_folder_path / f"{self.ModName}.py"
            linecache.checkcache(str(module_file))

        try:
            self.node_mod = import_module(mod_path)
            self._parse_existing_module()
        except ModuleNotFoundError:
            if self.log is None:
                print("No code generated yet, building from scratch")
            else:
                self.log.info("No code generated yet, building from scratch")

    def _parse_existing_module(self):
        """Parse an existing generated module to extract class definitions."""
        # Prevent circular import

        from secop_ophyd.SECoPDevices import (
            SECoPDevice,
            SECoPNodeDevice,
        )

        if self.node_mod is None:
            return

        modules = inspect.getmembers(self.node_mod)
        # Filter to only classes defined in this module, not imported ones
        class_members = [
            m
            for m in modules
            if inspect.isclass(m[1]) and m[1].__module__ == self.node_mod.__name__
        ]

        enum_classes = [m for m in class_members if issubclass(m[1], StrEnum)]
        node_classes = [m for m in class_members if issubclass(m[1], SECoPNodeDevice)]
        module_classes = [
            m
            for m in class_members
            if issubclass(m[1], SECoPDevice) and not issubclass(m[1], SECoPNodeDevice)
        ]

        for class_symbol, class_obj in enum_classes:
            self._parse_enum_class(class_symbol, class_obj)

        for class_symbol, class_obj in node_classes:
            self._parse_node_class(class_symbol, class_obj)

        for class_symbol, class_obj in module_classes:
            self._parse_module_class(class_symbol, class_obj)

    def _parse_node_class(self, class_symbol: str, class_obj: type):
        """Parse a node class from existing module.

        Args:
            class_symbol: Name of the class
            class_obj: The class object
        """
        # attrs = self._extract_attrs_from_source(inspect.getsource(class_obj))

        bases = [base.__name__ for base in class_obj.__bases__]

        # Extract description from docstring
        description = inspect.getdoc(class_obj) or ""

        _, properties, modules = self._get_attr_list(class_obj)

        node_cls = NodeClass(
            name=class_symbol,
            bases=bases,
            properties=properties,
            modules=modules,
            description=description,
        )
        self.node_classes.append(node_cls)

    def _extract_descriptions_from_source(self, class_obj: type) -> dict[str, str]:
        """Extract trailing comment descriptions from class source code.

        Uses ``ast`` to find class-level annotated attributes and ``tokenize`` to
        read actual Python comment tokens. This avoids false positives from ``#``
        inside strings and ignores non-attribute annotations.

        Args:
            class_obj: The class object to extract descriptions from

        Returns:
            Dictionary mapping attribute names to their descriptions
        """
        descriptions: dict[str, str] = {}
        try:
            source = textwrap.dedent(inspect.getsource(class_obj))
            source_lines = source.splitlines()
            module_ast = ast.parse(source)

            class_nodes = [
                node for node in module_ast.body if isinstance(node, ast.ClassDef)
            ]
            if not class_nodes:
                return descriptions

            class_node = class_nodes[0]

            comments_by_line: dict[int, list[str]] = {}
            for token_info in tokenize.generate_tokens(io.StringIO(source).readline):
                if token_info.type != tokenize.COMMENT:
                    continue

                comment_text = token_info.string[1:].lstrip().rstrip()
                comments_by_line.setdefault(token_info.start[0], []).append(
                    comment_text
                )

            for node in class_node.body:
                if not isinstance(node, ast.AnnAssign):
                    continue

                if not isinstance(node.target, ast.Name):
                    continue

                attr_name = node.target.id
                annotation_end_line = getattr(node, "end_lineno", node.lineno)
                description_lines: list[str] = []

                # Inline comment on the annotation line.
                description_lines.extend(comments_by_line.get(annotation_end_line, []))

                # Multiline trailing comment block directly below the annotation.
                next_line_no = annotation_end_line + 1
                while next_line_no <= len(source_lines):
                    stripped_line = source_lines[next_line_no - 1].lstrip()
                    if not stripped_line.startswith("#"):
                        break

                    description_lines.extend(comments_by_line.get(next_line_no, []))
                    next_line_no += 1

                description = "\n".join(description_lines).rstrip()
                if description:
                    descriptions[attr_name] = description
        except Exception as e:
            if self.log:
                self.log.debug(f"Could not extract descriptions from source: {e}")

        return descriptions

    def _normalize_description(self, description: str | None) -> str:
        """Normalize description text for generated comments.

        - Trim trailing whitespace/newlines
        - Preserve intentional internal newlines
        """
        if description is None:
            return ""

        normalized = description.rstrip()
        return normalized if normalized else ""

    def _get_attr_list(
        self, class_obj: type
    ) -> tuple[
        list[ParameterAttribute], list[PropertyAttribute], list[ModuleAttribute]
    ]:
        hints = get_type_hints(class_obj)
        # Get hints with Annotated for wrapping signals and backends
        extra_hints = get_type_hints(class_obj, include_extras=True)

        # Extract description comments from source code
        descriptions = self._extract_descriptions_from_source(class_obj)

        modules = []
        properties = []
        parameters = []

        for attr_name, annotation in hints.items():
            extras = getattr(extra_hints[attr_name], "__metadata__", ())

            origin = get_origin_class(annotation)

            if issubclass(origin, Signal):

                sig_type = annotation.__args__[0]
                # Get the module name
                module = sig_type.__module__

                type_param = (
                    sig_type.__name__ if module == "builtins" else sig_type.__name__
                )

                path_annotation = next(
                    (e for e in extras if isinstance(e, (ParameterType, PropertyType))),
                    None,
                )
                category = (
                    "property"
                    if isinstance(path_annotation, PropertyType)
                    else "parameter"
                )
                format_annotation = next(
                    (e for e in extras if isinstance(e, Format)), None
                )
                if format_annotation is not None:
                    format_annotation = f"Format.{format_annotation.name}"

                # Get description from comments
                description = descriptions.get(attr_name)

                match category:
                    case "property":
                        properties.append(
                            PropertyAttribute(
                                name=attr_name,
                                type=origin.__name__,
                                type_param=type_param,
                                path_annotation=str(path_annotation),
                            )
                        )
                    case "parameter":
                        parameters.append(
                            ParameterAttribute(
                                name=attr_name,
                                type=origin.__name__,
                                type_param=type_param,
                                description=description,
                                path_annotation=str(path_annotation),
                                format_annotation=format_annotation,
                            )
                        )

            if issubclass(origin, StandardReadable):
                modules.append(ModuleAttribute(name=attr_name, type=origin.__name__))

        return parameters, properties, modules

    def _parse_enum_class(self, class_symbol: str, class_obj: type):
        """Parse an enum class from existing module.

        Args:
            class_symbol: Name of the class
            class_obj: The class object

        """
        # Extract description from docstring
        description = inspect.getdoc(class_obj) or ""

        # Extract enum members from class attributes
        members = []

        for attr_name, attr_value in class_obj.__dict__.items():
            # Skip private/magic attributes and methods
            if attr_name.startswith("_") or callable(attr_value):
                continue

            # Create an EnumMember for each enum value
            # attr_name is the member name (e.g., "RAMP")
            # attr_value is the member value (e.g., "ramp")
            member = EnumMember(name=attr_name, value=attr_value, description=None)
            members.append(member)

        bases = [base.__name__ for base in class_obj.__bases__]

        # Create and return the EnumClass
        self.enum_classes.append(
            EnumClass(
                name=class_symbol,
                members=members,
                description=description,
                base_enum_class=bases[0] if bases else "StrictEnum",
            )
        )

    def _parse_module_class(self, class_symbol: str, class_obj: type):
        """Parse a module class from existing module.

        Args:
            class_symbol: Name of the class
            class_obj: The class object
        """
        # Extract attributes from source code to get proper type annotations

        parameters, properties, _ = self._get_attr_list(class_obj)

        methods = []
        for method_name, method in class_obj.__dict__.items():
            if callable(method) and not method_name.startswith("__"):
                method_source = inspect.getsource(method)
                description = self._extract_method_description(method_source)
                methods.append(
                    Method(method_name, description, inspect.signature(method))
                )

        bases = [base.__name__ for base in class_obj.__bases__]

        # Extract description from docstring
        description = inspect.getdoc(class_obj) or ""

        mod_enums: list[EnumClass] = []
        enums = {enum_class.name: enum_class for enum_class in self.enum_classes}

        for param in parameters:
            if param.type_param in enums:
                enum_class = enums[param.type_param]
                if enum_class not in mod_enums:
                    mod_enums.append(enum_class)

        mod_cls = ModuleClass(
            name=class_symbol,
            bases=bases,
            parameters=parameters,
            properties=properties,
            methods=methods,
            description=description,
            enums=mod_enums,
        )
        self.module_classes.append(mod_cls)

    def _extract_method_description(self, method_source: str) -> str:
        """Extract description from method docstring.

        Args:
            method_source: Source code of the method

        Returns:
            Description string
        """
        match = re.search(r"\s*def\s+\w+\s*\(.*\).*:\s*", method_source)
        if match:
            function_body = method_source[match.end() :]
            description_list = function_body.split('"""', 2)
            if len(description_list) > 1:
                return description_list[1]
        return ""

    def add_import(self, module: str, class_str: str | None = None):
        """Add an import to the import dictionary.

        Args:
            module: Python module to import from
            class_str: Class/symbol to import. If None or empty, imports the module
            directly.
        """
        if class_str is None or class_str == "":
            # For module-only imports (import module), use None as value
            if module not in self.imports:
                self.imports[module] = None
        else:
            existing = self.imports.get(module)
            if existing is None:
                # Convert from module-only import to specific imports
                self.imports[module] = {class_str}
            elif isinstance(existing, set):
                existing.add(class_str)
            else:
                self.imports[module] = {class_str}

    def add_mod_class(
        self,
        module_cls: str,
        bases: list[str],
        parameters: list[ParameterAttribute],
        properties: list[PropertyAttribute],
        cmd_plans: list[Method],
        description: str = "",
        enum_classes: list[EnumClass] | None = None,
    ):
        """Add a module class to be generated.

        Args:
            module_cls: Name of the module class
            bases: Base classes
            parameters: List of parameter attributes
            properties: List of property attributes
            cmd_plans: List of method definitions
            description: Optional class description
        """
        # Check if class already exists (loaded from file)
        existing_class = next(
            (cls for cls in self.module_classes if cls.name == module_cls), None
        )
        if existing_class:
            # Class already exists - merge enums if provided
            if enum_classes:
                existing_class.enums.extend(enum_classes)
                if self.log:
                    self.log.info(
                        f"Module class {module_cls} already exists, "
                        f"merged {len(enum_classes)} enum(s)"
                    )
            return

        mod_cls = ModuleClass(
            name=module_cls,
            bases=bases,
            parameters=parameters,
            properties=properties,
            methods=cmd_plans,
            description=description,
            enums=enum_classes or [],
        )
        self.module_classes.append(mod_cls)

    def add_node_class(
        self,
        node_cls: str,
        bases: list[str],
        properties: list[PropertyAttribute],
        modules: list[ModuleAttribute],
        description: str = "",
    ):
        """Add a node class to be generated.

        Args:
            node_cls: Name of the node class
            bases: Base classes
            attrs: List of attribute tuples. Supported formats:
                   - (name, type)
                   - (name, type, type_param)
                   - (name, type, type_param, description, category)
        """
        # Check if class already exists (loaded from file)
        existing_class = next(
            (cls for cls in self.node_classes if cls.name == node_cls), None
        )
        if existing_class:
            # Class already exists, skip adding it
            if self.log:
                self.log.info(f"Node class {node_cls} already exists, skipping")
            return

        node_class = NodeClass(
            name=node_cls,
            bases=bases,
            properties=properties,
            modules=modules,
            description=description,
        )
        self.node_classes.append(node_class)

    def _parse_command_signature(
        self, cmd_name: str, datainfo: dict, description: str
    ) -> Method:
        """Parse command datainfo to create Method signature.

        Args:
            cmd_name: Name of the command
            datainfo: Command datainfo with argument/result types
            description: Command description

        Returns:
            Method object with signature
        """
        # Extract argument and result types
        arg_type = datainfo.get("argument")

        # Create a basic signature object
        sig = Signature.from_callable(lambda self, wait_for_idle=False: None)
        if arg_type is not None:
            sig = Signature.from_callable(lambda self, arg, wait_for_idle=False: None)

        return Method(cmd_name=cmd_name, description=description, cmd_sign=sig)

    def from_json_describe(self, json_data: str | dict):
        """Generate classes from a SECoP JSON describe message.

        Args:
            json_data: JSON string or dict containing SECoP describe data
        """
        # Parse JSON if string
        if isinstance(json_data, str):
            describe_data = json.loads(json_data)
        else:
            describe_data = json_data

        modules: dict[str, dict] = describe_data.get("modules", {})
        node_properties = {k: v for k, v in describe_data.items() if k != "modules"}

        # Parse modules
        node_module_attrs: list[ModuleAttribute] = []
        node_property_attrs: list[PropertyAttribute] = []

        for modname, moddescr in modules.items():
            #  separate accessibles into command and parameters
            parameters = {}
            commands = {}
            accessibles = moddescr["accessibles"]
            for aname, aentry in accessibles.items():
                iname = internalize_name(aname)
                datatype = get_datatype(aentry["datainfo"], iname)

                aentry = dict(aentry, datatype=datatype)

                if datatype.IS_COMMAND:
                    commands[iname] = aentry
                else:
                    parameters[iname] = aentry

            properties = {k: v for k, v in moddescr.items() if k != "accessibles"}

            # Add module class (highest secop interface class) that the actual
            # module class is derived from
            secop_ophyd_modclass = class_from_interface(properties)
            module_bases = [secop_ophyd_modclass.__name__]

            # Add the module class, use self reported "implementation" module property,
            # if not present use the module name
            module_class = modname
            if properties.get("implementation", ""):
                module_class = properties.get("implementation", "").split(".").pop()

            module_class_list = (
                module_class.replace(" ", "_").replace("-", "_").split("_")
            )

            module_class = "".join(word.capitalize() for word in module_class_list)

            # Module enum classes
            module_enum_classes = []

            # Prepare attributes

            # Module Commands
            command_plans = []

            for command, command_data in commands.items():
                # Stop is already an ophyd native operation
                if command == "stop":
                    continue

                argument = command_data["datainfo"].get("argument")
                result = command_data["datainfo"].get("result")

                description: str = ""
                description += f"{command_data['description']}\n"

                if argument:
                    description += (
                        f"       argument: {command_data['datainfo'].get('argument')}\n"
                    )
                if result:
                    description += (
                        f"       result: {command_data['datainfo'].get('result')}"
                    )

                def command_plan(self, arg, wait_for_idle: bool = False):
                    pass

                def command_plan_no_arg(self, wait_for_idle: bool = False):
                    pass

                plan = Method(
                    cmd_name=command,
                    description=description,
                    cmd_sign=inspect.signature(
                        command_plan if argument else command_plan_no_arg
                    ),
                )

                command_plans.append(plan)

            mod_parameters: list[ParameterAttribute] = []

            for param_name, param_data in parameters.items():

                descr = self._normalize_description(param_data.get("description", ""))
                unit = param_data["datainfo"].get("unit")

                if unit:
                    param_descr = (
                        f"{descr}; Unit: ({unit})" if descr else f"Unit: ({unit})"
                    )
                else:
                    param_descr = descr
                signal_base = SignalR if param_data["readonly"] else SignalRW

                format = None

                # infer format from parameter property
                match param_data.get("_signal_format", None):
                    case "HINTED_SIGNAL":
                        format = Format.HINTED_SIGNAL
                    case "HINTED_UNCACHED_SIGNAL":
                        format = Format.HINTED_UNCACHED_SIGNAL
                    case "UNCACHED_SIGNAL":
                        format = Format.UNCACHED_SIGNAL
                    case _:
                        format = None

                # depending on the Interface class other parameter need to be declared
                # as readsignals as well
                if param_name in secop_ophyd_modclass.hinted_signals:
                    format = format or Format.HINTED_SIGNAL

                # Remove "StandardReadable" prefix from format for cleaner annotation
                format = (
                    str(format).removeprefix("StandardReadable") if format else None
                )

                datainfo = param_data.get("datainfo", {})

                # infer the ophyd type from secop datatype
                type_param = get_type_param(param_data["datatype"])

                # Handle StrictEnum types - generate enum class
                if type_param and "StrictEnum" in type_param:
                    # Generate unique enum class name:
                    # ModuleClass + ParamName + Enum
                    param_name_list = (
                        param_name.replace(" ", "_").replace("-", "_").split("_")
                    )

                    param_name_camel = "".join(
                        word.capitalize() for word in param_name_list
                    )

                    enum_class_name = f"{module_class}_{param_name_camel}_Enum"

                    # Extract enum members from datainfo
                    enum_members_dict = datainfo.get("members", {})
                    if enum_members_dict:
                        from secop_ophyd.GenNodeCode import EnumClass, EnumMember

                        enum_members = []
                        for member_value, _ in enum_members_dict.items():
                            # Convert member name to Python identifier
                            python_name = secop_enum_name_to_python(member_value)
                            enum_members.append(
                                EnumMember(
                                    name=python_name,
                                    value=member_value,
                                    description=None,
                                )
                            )

                        # Create enum class definition
                        enum_descr = f"{param_name} enum for `{module_class}`."

                        enum_cls = EnumClass(
                            name=enum_class_name,
                            members=enum_members,
                            description=enum_descr,
                        )

                        module_enum_classes.append(enum_cls)

                        # Use the specific enum class name instead of generic
                        # StrictEnum
                        type_param = enum_class_name

                # Default format for parameters is CONFIG_SIGNAL

                mod_parameters.append(
                    ParameterAttribute(
                        name=param_name,
                        type=signal_base.__name__,
                        type_param=type_param,
                        description=param_descr,
                        path_annotation=str(ParameterType()),
                        format_annotation=format,
                    )
                )

            # Module properties
            module_properties: list[PropertyAttribute] = []

            # Process module properties
            for prop_name, property_value in properties.items():

                if prop_name in IGNORED_PROPS:
                    continue

                type_param = get_type_param(secop_dtype_obj_from_json(property_value))

                module_properties.append(
                    PropertyAttribute(
                        name=prop_name,
                        type=SignalR.__name__,
                        type_param=type_param,
                        path_annotation=str(PropertyType()),
                    )
                )

            self.add_mod_class(
                module_cls=module_class,
                bases=module_bases,
                parameters=mod_parameters,
                properties=module_properties,
                cmd_plans=command_plans,
                description=properties.get("description", ""),
                enum_classes=module_enum_classes,
            )

            # Add to node attributes
            # Type the None explicitly as str | None to match other entries
            node_module_attrs.append(ModuleAttribute(name=modname, type=module_class))

        # Process module properties
        for prop_name, property_value in node_properties.items():
            type_param = get_type_param(secop_dtype_obj_from_json(property_value))

            # Generate PropPath annotation for node-level properties

            node_property_attrs.append(
                PropertyAttribute(
                    name=prop_name,
                    type=SignalR.__name__,
                    type_param=type_param,
                    path_annotation=str(PropertyType()),
                )
            )

        # Add node class
        node_bases = ["SECoPNodeDevice"]

        equipment_id: str = node_properties["equipment_id"]

        # format node class accordingly
        node_class_name = equipment_id.replace(".", "_").replace("-", "_").capitalize()

        self.add_node_class(
            node_cls=node_class_name,
            bases=node_bases,
            modules=node_module_attrs,
            properties=node_property_attrs,
            description=node_properties.get("description", ""),
        )

        # Add required imports
        self.add_import("secop_ophyd.SECoPDevices", "SECoPNodeDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPBaseDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPCommunicatorDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPReadableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPWritableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPMoveableDevice")
        self.add_import("secop_ophyd.SECoPDevices", "SECoPTriggerableDevice")

    def generate_code(self) -> str:
        """Generate Python code using Jinja2 template.

        Returns:
            Generated Python code as string
        """
        template = self.jinja_env.get_template("generated_classes.py.jinja2")

        # Prepare template context
        context = {
            "imports": self.imports,
            "module_classes": self.module_classes,
            "node_classes": self.node_classes,
            "enum_classes": self._collect_all_enums(),
            "inline_comment_threshold": self.inline_comment_threshold,
            "comment_wrap_width": self.comment_wrap_width,
        }

        # Render template
        code = template.render(**context)

        # Remove unused imports with autoflake
        try:

            code = autoflake.fix_code(
                code,
                remove_all_unused_imports=True,
                remove_unused_variables=False,
                remove_duplicate_keys=True,
            )
        except Exception as e:
            if self.log:
                self.log.warning(f"Autoflake processing failed: {e}")
            else:
                print(f"Warning: Autoflake processing failed: {e}")

        # Format with Black
        try:
            code = black.format_str(code, mode=black.Mode(line_length=200))
        except Exception as e:
            if self.log:
                self.log.warning(f"Black formatting failed: {e}")
            else:
                print(f"Warning: Black formatting failed: {e}")

        return code

    def _collect_all_enums(self) -> list[EnumClass]:
        """Collect and merge enum definitions from all module classes.

        When multiple module classes have enums with the same base name but different
        members, they are merged into a single SupersetEnum containing the union of all
        members.

        Returns:
            List of deduplicated EnumClass definitions
        """
        from collections import defaultdict

        # Group enums by their base name (ModuleClass + ParamName + Enum)
        # We need to track which module classes use each enum
        enum_groups = defaultdict(list)  # base_name -> [(module_class, enum)]

        for mod_cls in self.module_classes:
            for enum in mod_cls.enums:
                # Extract base enum name by removing module class prefix
                # e.g., "MassflowController1Gastype_selectEnum" -> need module class
                # name
                enum_groups[enum.name].append((mod_cls.name, enum))

        # Process each enum group
        merged_enums = []
        for enum_name, enum_list in enum_groups.items():
            if len(enum_list) == 1:
                # Single enum definition - use StrictEnum
                _, enum = enum_list[0]

                # an already merged enum read in from a file
                if enum.base_enum_class != "StrictEnum":
                    merged_enums.append(enum)
                    continue

                enum.base_enum_class = "StrictEnum"
                merged_enums.append(enum)
            else:
                # Multiple definitions - need to check if members are identical
                member_sets = [
                    frozenset((m.name, m.value) for m in enum.members)
                    for _, enum in enum_list
                ]

                if len(set(member_sets)) == 1:
                    # All enums have identical members - use StrictEnum
                    _, enum = enum_list[0]
                    enum.base_enum_class = "StrictEnum"
                    merged_enums.append(enum)
                else:
                    # Different members - merge into SupersetEnum
                    all_members_dict = {}  # (name, value) -> EnumMember

                    for _, enum in enum_list:
                        for member in enum.members:
                            key = (member.name, member.value)
                            if key not in all_members_dict:
                                all_members_dict[key] = member

                    # Create merged enum with all unique members
                    _, base_enum = enum_list[0]
                    merged_enum = EnumClass(
                        name=enum_name,
                        members=list(all_members_dict.values()),
                        description=base_enum.description,
                        base_enum_class="SupersetEnum",
                    )
                    merged_enums.append(merged_enum)

        return merged_enums

    def write_gen_node_class_file(self):
        """Generate and write the class file to disk."""
        code = self.generate_code()

        # Determine file path
        if self.module_folder_path is None:
            filep = Path(f"{self.ModName}.py")
        else:
            filep = self.module_folder_path / f"{self.ModName}.py"

        # Write to file
        with open(filep, "w") as file:
            file.write(code)

        if self.log:
            self.log.info(f"Generated class file: {filep}")
        else:
            print(f"Generated class file: {filep}")

        # Reload the module
        if self.node_mod is not None:
            reload(self.node_mod)


def get_type_param(secop_dtype: DataType) -> str | None:
    sig_type = SECoPdtype(secop_dtype).np_datatype

    # Get the module name
    module = sig_type.__module__

    # For builtins, just return the name without module prefix
    if module == "builtins":
        return sig_type.__name__

    return sig_type.__name__


def get_type_prop(prop_value) -> str | None:
    secop_dtype: DataType = secop_dtype_obj_from_json(prop_value)
    return get_type_param(secop_dtype)



"""Simple test to verify GenNodeCode refactoring works."""

import inspect
import sys
from pathlib import Path

from ophyd_async.core import SignalR, init_devices

from secop_ophyd.GenNodeCode import (
    GenNodeCode,
    Method,
    ModuleAttribute,
    ModuleClass,
    NodeClass,
    ParameterAttribute,
    PropertyAttribute,
)
from secop_ophyd.SECoPDevices import ParameterType, PropertyType, SECoPNodeDevice

# Add src to path
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))


class _DescriptionParseSample:
    count: int  # count comment
    label: str = "value # not a comment"
    # label continuation

    def helper(self):
        local: int  # must not be parsed  # noqa: F842


def test_extract_descriptions_from_source_is_token_safe():
    """Ensure parser only reads real comments and ignores '#' inside strings."""
    gen_code = GenNodeCode(log=None)
    descriptions = gen_code._extract_descriptions_from_source(_DescriptionParseSample)

    assert descriptions["count"] == "count comment"
    assert descriptions["label"] == "label continuation"
    assert "local" not in descriptions


def test_generated_command_methods_are_concrete(tmp_path: Path):
    """Generated command wrappers must be concrete so devices are instantiable."""
    gen_code = GenNodeCode(path=str(tmp_path), log=None)

    def command_plan_no_arg(self, wait_for_idle: bool = False):
        pass

    gen_code.add_mod_class(
        module_cls="CommandTestModule",
        bases=["SECoPReadableDevice"],
        parameters=[],
        properties=[],
        cmd_plans=[
            Method(
                cmd_name="factory_reset",
                description="Reset module",
                cmd_sign=inspect.signature(command_plan_no_arg),
            )
        ],
        description="test module",
    )

    generated_code = gen_code.generate_code()

    assert "def factory_reset" in generated_code
    assert "@abstractmethod" not in generated_code
    assert "raise RuntimeError(" in generated_code


def test_basic_functionality(clean_generated_file):
    """Test basic GenNodeCode functionality."""
    print("Testing GenNodeCode refactored implementation...")

    # Create instance
    gen_code = GenNodeCode(path=str(clean_generated_file), log=None)

    # Add some imports
    gen_code.add_import("ophyd_async.core", "Device")

    # Create a simple method
    from inspect import signature

    def sample_method(self, value: int) -> str:
        """Sample method description"""
        return str(value)

    # Method can be created in the old way (backward compatible)
    method = Method(
        cmd_name="sample_command",
        description="Sample command for testing",
        cmd_sign=signature(sample_method),
    )

    # Add a module class
    gen_code.add_mod_class(
        module_cls="TestModule",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="temperature",
                type="SignalR",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="pressure",
                type="SignalR",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="count",
                type="SignalRW",
                type_param="int",
                path_annotation=str(ParameterType()),
            ),
        ],
        properties=[],
        cmd_plans=[method],
        description="Test module class",
    )

    # Add a node class
    gen_code.add_node_class(
        node_cls="TestNode",
        bases=["SECoPNodeDevice"],
        modules=[
            ModuleAttribute(name="module1", type="TestModule"),
        ],
        properties=[
            PropertyAttribute(
                name="status",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
        ],
        description="Test node class",
    )

    # Generate code
    code = gen_code.generate_code()

    gen_code.write_gen_node_class_file()

    print("\n" + "=" * 60)
    print("Generated Code:")
    print("=" * 60)
    print(code)
    print("=" * 60)

    # Verify code contains expected elements
    assert "class TestModule(SECoPDevice):" in code
    assert "temperature: A[SignalR[float], ParamT()]" in code
    assert "count: A[SignalRW[int], ParamT()]" in code
    assert "class TestNode(SECoPNodeDevice):" in code
    assert "module1: TestModule" in code
    assert "status: A[SignalR[str], PropT()]" in code
    assert "def sample_command" in code

    print("\n✓ All basic tests passed!")


def test_dataclasses():
    """Test the new dataclasses."""
    print("\nTesting dataclasses...")

    # Test ParameterAttribute
    param_attr = ParameterAttribute(name="test_param", type="SignalR")
    assert param_attr.name == "test_param"
    assert param_attr.type == "SignalR"

    # Test PropertyAttribute
    prop_attr = PropertyAttribute(name="test_prop", type="SignalR")
    assert prop_attr.name == "test_prop"
    assert prop_attr.type == "SignalR"

    # Test ModuleAttribute
    mod_attr = ModuleAttribute(name="test_mod", type="TestModule")
    assert mod_attr.name == "test_mod"
    assert mod_attr.type == "TestModule"

    # Test ModuleClass
    mod_cls = ModuleClass(
        name="TestMod",
        bases=["Device"],
        parameters=[param_attr],
        properties=[prop_attr],
        methods=[],
    )
    assert mod_cls.name == "TestMod"
    assert len(mod_cls.parameters) == 1
    assert len(mod_cls.properties) == 1

    # Test NodeClass
    node_cls = NodeClass(
        name="TestNode", bases=["Device"], modules=[mod_attr], properties=[prop_attr]
    )
    assert node_cls.name == "TestNode"

    print("✓ Dataclass tests passed!")


def test_subsequent_node_generation(clean_generated_file):
    """Test generating code for two nodes sequentially, appending to the same file.

    Tests that:
    - First: Generate NodeA with modules Type1 and Type2, write to file
    - Second: Load existing file, add NodeB with Type1 (shared) and Type3 (new)
    - Type1 should appear only once in the final file (not duplicated)
    - All classes (Type1, Type2, Type3, NodeA, NodeB) are in the final file
    """

    from inspect import signature

    # ===== STEP 1: Generate and write first node (NodeA) =====

    gen_code1 = GenNodeCode(path=str(clean_generated_file), log=None)

    # Create sample methods
    def type1_command(self, value: float) -> float:
        """Type1 command"""
        return value * 2.0

    method_type1 = Method(
        cmd_name="type1_cmd",
        description="Type1 command",
        cmd_sign=signature(type1_command),
    )

    def type2_command(self, mode: str) -> str:
        """Type2 command"""
        return f"Mode: {mode}"

    method_type2 = Method(
        cmd_name="type2_cmd",
        description="Type2 command",
        cmd_sign=signature(type2_command),
    )

    # Add module class Type1 (will be shared)
    gen_code1.add_mod_class(
        module_cls="Type1",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="temperature",
                type="SignalR",
                type_param="float",
                description="this has to be in the final output",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="setpoint",
                type="SignalRW",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
        ],
        properties=[
            PropertyAttribute(
                name="description",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
            PropertyAttribute(
                name="interface_classes",
                type="SignalR",
                type_param="int",
                path_annotation=str(PropertyType()),
            ),
        ],
        cmd_plans=[method_type1],
        description="Type1 module - shared between nodes",
    )

    # Add module class Type2 (only in nodeA)
    gen_code1.add_mod_class(
        module_cls="Type2",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="pressure",
                type="SignalR",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="mode",
                type="SignalRW",
                type_param="str",
                path_annotation=str(ParameterType()),
            ),
        ],
        properties=[
            PropertyAttribute(
                name="implementation",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
        ],
        cmd_plans=[method_type2],
        description="Type2 module - only in nodeA",
    )

    # Add nodeA
    gen_code1.add_node_class(
        node_cls="NodeA",
        bases=["SECoPNodeDevice"],
        modules=[
            ModuleAttribute(name="modA", type="Type1"),
            ModuleAttribute(name="modB", type="Type2"),
        ],
        properties=[
            PropertyAttribute(
                name="status",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
        ],
        description="NodeA with Type1 and Type2 modules",
    )

    # Generate and write first node
    code1 = gen_code1.generate_code()
    gen_code1.write_gen_node_class_file()

    # Verify first generation
    assert "class Type1(SECoPDevice):" in code1
    assert "class Type2(SECoPDevice):" in code1
    assert "class NodeA(SECoPNodeDevice):" in code1
    assert "modA: Type1" in code1
    assert "modB: Type2" in code1

    # ===== STEP 2: Load existing file and add second node (NodeB) =====

    gen_code2 = GenNodeCode(path=str(clean_generated_file), log=None)

    # Add necessary imports again
    gen_code2.add_import("secop_ophyd.SECoPDevices", "SECoPDevice")
    gen_code2.add_import("secop_ophyd.SECoPDevices", "SECoPNodeDevice")

    # Create method for Type3
    def type3_command(self, count: int) -> int:
        """Type3 command"""
        return count + 1

    method_type3 = Method(
        cmd_name="type3_cmd",
        description="Type3 command",
        cmd_sign=signature(type3_command),
    )

    # Add Type1 again - GenNodeCode should detect it already exists
    gen_code2.add_mod_class(
        module_cls="Type1",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="temperature",
                type="SignalR",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="setpoint",
                type="SignalRW",
                type_param="float",
                path_annotation=str(ParameterType()),
            ),
        ],
        properties=[
            PropertyAttribute(
                name="description",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
            PropertyAttribute(
                name="interface_classes",
                type="SignalR",
                type_param="list",
                path_annotation=str(PropertyType()),
            ),
        ],
        cmd_plans=[method_type1],
        description="Type1 module - shared between nodes",
    )

    # Add module class Type3 (only in nodeB)
    gen_code2.add_mod_class(
        module_cls="Type3",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="count",
                type="SignalRW",
                type_param="int",
                description="this is a description",
                path_annotation=str(ParameterType()),
            ),
            ParameterAttribute(
                name="enabled",
                type="SignalR",
                type_param="bool",
                path_annotation=str(ParameterType()),
            ),
        ],
        properties=[
            PropertyAttribute(
                name="group",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
        ],
        cmd_plans=[method_type3],
        description="Type3 module - only in nodeB",
    )

    # Add nodeB
    gen_code2.add_node_class(
        node_cls="NodeB",
        bases=["SECoPNodeDevice"],
        modules=[
            ModuleAttribute(name="modA", type="Type1"),
            ModuleAttribute(name="modB", type="Type3"),
        ],
        properties=[
            PropertyAttribute(
                name="name",
                type="SignalR",
                type_param="str",
                path_annotation=str(PropertyType()),
            ),
        ],
        description="NodeB with Type1 and Type3 modules",
    )

    # Generate and write second node (appends to the file)
    code2 = gen_code2.generate_code()
    gen_code2.write_gen_node_class_file()

    # ===== VERIFICATION =====
    # Verify that Type1 appears only once in the final code
    type1_count = code2.count("class Type1(SECoPDevice):")

    assert (
        type1_count == 1
    ), f"Type1 should appear exactly once, but appears {type1_count} times"

    # Verify all module classes are present
    assert "class Type1(SECoPDevice):" in code2
    assert "class Type2(SECoPDevice):" in code2
    assert "class Type3(SECoPDevice):" in code2

    # Verify both node classes are present
    assert "class NodeA(SECoPNodeDevice):" in code2
    assert "class NodeB(SECoPNodeDevice):" in code2

    # Verify all methods are present
    assert "def type1_cmd" in code2
    assert "def type2_cmd" in code2
    assert "def type3_cmd" in code2

    # Verify section comments are present
    assert "# Module Properties" in code2
    assert "# Module Parameters" in code2

    # Verify that descriptive  comments are preserved in generated code
    assert "# this is a description" in code2
    assert "# this has to be in the final output" in code2


async def test_gen_cryo_node(
    clean_generated_file, cryo_sim, cryo_node_no_re: SECoPNodeDevice
):
    """Test generating code for a real SECoP node."""

    cryo_node_no_re.class_from_instance(clean_generated_file)

    from tests.testgen.genNodeClass import Cryo_7_frappy_demo  # type: ignore

    async with init_devices():
        cryo_gen_code = Cryo_7_frappy_demo(sec_node_uri="localhost:10769")

    cryo_val = await cryo_gen_code.read()

    # target and value shoule be present in readback, since they are read signals with
    # HINTED Format --> this is tested to verify that the correct annotations are
    # generated and interpreted in the generated code
    val_name = cryo_gen_code.cryo.value.name
    target_name = cryo_gen_code.cryo.target.name
    read_val = cryo_val[val_name].get("value")
    read_target = cryo_val[target_name].get("value")

    print(cryo_val)

    assert read_val is not None
    assert read_val > 5

    assert read_target is not None
    assert read_target == 10


async def test_gen_cryo_status_not_in_cfg(
    clean_generated_file, cryo_sim, cryo_node_no_re: SECoPNodeDevice
):
    """Test that Status signal is not marked as configuration signal but is still
    instantiated."""

    cryo_node_no_re.class_from_instance(clean_generated_file)

    cryo_cfg = await cryo_node_no_re.read_configuration()
    cryo_reading = await cryo_node_no_re.read()

    print(cryo_reading)

    assert hasattr(cryo_node_no_re.cryo, "status")
    assert isinstance(cryo_node_no_re.cryo.status, SignalR)

    stat_name = cryo_node_no_re.cryo.status.name

    assert (
        cryo_cfg.get(stat_name) is None
    ), "Status signal should not be in configuration"
    assert cryo_reading.get(stat_name) is None, "Status signal should be readable"

    # check if status signal is working
    status_reding = await cryo_node_no_re.cryo.status.read()

    assert status_reding.get(stat_name) is not None, "Status signal should be readable"

    # Import generated class
    from tests.testgen.genNodeClass import Cryo_7_frappy_demo  # type: ignore

    async with init_devices():
        cryo_gen_code = Cryo_7_frappy_demo(sec_node_uri="localhost:10769")

    # Status signal should still be present and functional in the generated code, even
    # though it's not in the configuration
    assert hasattr(cryo_gen_code.cryo, "status")
    assert isinstance(cryo_gen_code.cryo.status, SignalR)

    cryo_cfg = await cryo_gen_code.read_configuration()
    cryo_reading = await cryo_gen_code.read()

    print(cryo_reading)

    stat_name = cryo_gen_code.cryo.status.name

    assert (
        cryo_cfg.get(stat_name) is None
    ), "Status signal should not be in configuration"
    assert cryo_reading.get(stat_name) is None, "Status signal should be readable"

    # check if status signal is working
    status_reding = await cryo_gen_code.cryo.status.read()

    assert status_reding.get(stat_name) is not None, "Status signal should be readable"


async def test_gen_real_node(
    clean_generated_file,
    nested_struct_sim,
    nested_node_no_re: SECoPNodeDevice,  # noqa: N803
):

    nested_node_no_re.class_from_instance(clean_generated_file)

    # Read the generated file and verify its contents
    gen_file = clean_generated_file / "genNodeClass.py"
    assert gen_file.exists(), "Generated file should exist"

    generated_code = gen_file.read_text()

    # ===== Assertions for generated command plans =====
    # The ophy_struct module has a test_cmd command
    assert "def test_cmd" in generated_code, "test_cmd plan should be generated"
    assert (
        "@abstractmethod" not in generated_code
    ), "Command methods should be concrete so generated classes are instantiable"

    # ===== Assertions for generated enum classes =====
    # Enum classes should be generated for enum parameters
    # The gas_type parameter in enum1/enum2 modules should generate enum classes
    assert (
        "class TestEnum_GasType_Enum(SupersetEnum):" in generated_code
    ), "Enum class for gas_type should be generated"

    # Verify enum members are present
    # gas_type enums should have AR, N2, H2 (and CO2 for enum2)
    assert "AR" in generated_code, "AR enum member should be present"
    assert "N2" in generated_code, "N2 enum member should be present"
    assert "H2" in generated_code, "H2 enum member should be present"
    assert "CO2" in generated_code, "CO2 enum member should be present"

    # Verify SupersetEnum import
    assert (
        "from enum import Enum" in generated_code or "SupersetEnum" in generated_code
    ), "Enum import should be present"


async def test_subsequent_real_nodes_with_enum(
    clean_generated_file,
    cryo_sim,
    cryo_node_no_re: SECoPNodeDevice,
    nested_struct_sim,
    nested_node_no_re: SECoPNodeDevice,
):

    nested_node_no_re.class_from_instance(clean_generated_file)

    # Read the generated file and verify its contents
    gen_file = clean_generated_file / "genNodeClass.py"
    assert gen_file.exists(), "Generated file should exist"

    generated_code = gen_file.read_text()

    # ===== Assertions for generated enum classes =====
    cls = [
        "class TestEnum_GasType_Enum(SupersetEnum):",
        "class TestModStr(SECoPReadableDevice):",
        "class OphydTestPrimitiveArrays(SECoPReadableDevice):",
        "class TestEnum(SECoPReadableDevice):",
        "class TestNdArrays(SECoPReadableDevice):",
        "class TestStructOfArrays(SECoPReadableDevice):",
        "class Ophyd_secop_frappy_demo(SECoPNodeDevice):",
    ]
    for classs_str in cls:
        assert classs_str in generated_code

    cryo_node_no_re.class_from_instance(clean_generated_file)

    # Read the generated file and verify its contents
    gen_file = clean_generated_file / "genNodeClass.py"
    assert gen_file.exists(), "Generated file should exist"

    generated_code = gen_file.read_text()

    # ===== Assertions for generated enum classes =====

    cls = [
        "class TestEnum_GasType_Enum(SupersetEnum):",
        "class TestModStr(SECoPReadableDevice):",
        "class OphydTestPrimitiveArrays(SECoPReadableDevice):",
        "class TestEnum(SECoPReadableDevice):",
        "class TestNdArrays(SECoPReadableDevice):",
        "class TestStructOfArrays(SECoPReadableDevice):",
        "class Ophyd_secop_frappy_demo(SECoPNodeDevice):",
        "class Cryo_7_frappy_demo(SECoPNodeDevice):",
        "class Cryostat(SECoPMoveableDevice):",
        "class Cryostat_Mode_Enum(StrictEnum):",
    ]
    for classs_str in cls:
        assert classs_str in generated_code


def test_gen_shall_mass_spec_node(
    clean_generated_file, mass_spectrometer_description: str
):
    """Test generating code for the SHALL mass spectrometer node using a
    real description."""

    gen_code = GenNodeCode(path=str(clean_generated_file))

    gen_code.from_json_describe(mass_spectrometer_description)

    gen_code.write_gen_node_class_file()

    gen_file = clean_generated_file / "genNodeClass.py"
    assert gen_file.exists(), "Generated file should exist"

    generated_code = gen_file.read_text()

    # Trailing newlines in source descriptions should not produce broken split comments
    assert "\n# ; Unit: (V)" not in generated_code
    assert "\n#  ; Unit: (%)" not in generated_code

    # Intentionally multiline descriptions should be rendered as multiline comments
    assert "mid_descriptor: A[SignalRW[ndarray], ParamT()]" in generated_code
    assert "#           Example:" in generated_code
    assert "#             {" in generated_code
    assert "#               mass:    [12,15,28,75]," in generated_code
    assert "#               device:  [FARADAY,SEM,SEM,SEM]" in generated_code

    # Long descriptions should be rendered fully below the declaration and wrapped
    assert "resolution: A[SignalR[float], ParamT()]\n" in generated_code
    assert (
        "#  The high mass peak width/valley adjustment used during set up and"
        in generated_code
    )
    assert (
        "# low masses and should be adjusted in conjunction with the Delta-M."
        in generated_code
    )

    # Reparse generated code and verify multiline comments survive round-trip generation
    roundtrip_gen = GenNodeCode(path=str(clean_generated_file))
    roundtrip_code = roundtrip_gen.generate_code()

    assert "mid_descriptor: A[SignalRW[ndarray], ParamT()]" in roundtrip_code
    assert "Example:" in roundtrip_code
    assert "\n# ; Unit: (V)" not in roundtrip_code
    assert "resolution: A[SignalR[float], ParamT()]\n" in roundtrip_code


def test_gen_shall_mass_spec_node_no_impl(
    clean_generated_file, mass_spectrometer_description_no_impl: str
):
    """Test generating code for the SHALL mass spectrometer node using a
    real description."""

    gen_code = GenNodeCode(path=str(clean_generated_file))

    gen_code.from_json_describe(mass_spectrometer_description_no_impl)

    gen_code.write_gen_node_class_file()



# mypy: disable-error-code="attr-defined"
import numpy as np
import pytest
from frappy.client import CacheItem
from frappy.datatypes import (
    ArrayOf,
    EnumType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from frappy.lib.enum import EnumMember

from secop_ophyd.util import SECoPdtype, SECoPReading

RAGGED = True
REGULAR = False


@pytest.mark.parametrize(
    "start_dtype,data,expected_shape,expected_update_shape",
    [
        pytest.param(
            ArrayOf(FloatRange(), minlen=0, maxlen=100),
            [0.1, 0.1],
            [100],
            [2],
            id="Ragged 1D",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange()), minlen=0, maxlen=100),
            [[10, 10], [10, 10]],
            [100, 100],  # maxlen defaults to 100 if not given
            [2, 2],
            id="Ragged 2D",
        ),
        pytest.param(
            ArrayOf(ArrayOf(ArrayOf(FloatRange())), minlen=0, maxlen=99),
            [[[10, 10], [10, 10]], [[10, 10], [10, 10]]],
            [99, 100, 100],
            [2, 2, 2],
            id="Ragged 3D",
        ),
        pytest.param(
            ArrayOf(
                ArrayOf(ArrayOf(FloatRange(), minlen=2, maxlen=2), minlen=2, maxlen=2),
                minlen=2,
                maxlen=2,
            ),
            [[[10, 10], [10, 10]], [[10, 10], [10, 10]]],
            [2, 2, 2],
            [2, 2, 2],
            id="Regular 3D",
        ),
        pytest.param(
            ArrayOf(TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType())),
            [([1, 2, 3], "blah")],
            [100],
            [1],
            id="Complex Ragged 1D",
        ),
        pytest.param(
            ArrayOf(
                ArrayOf(
                    TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType())
                )
            ),
            [
                [([1, 2, 3], "blah"), ([1, 2, 3], "blah")],
                [([1, 2, 3], "blah"), ([1, 2, 3], "blah")],
            ],
            [100, 100],
            [2, 2],
            id="Complex Ragged 2D",
        ),
        pytest.param(
            TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType()),
            ([1, 2, 3], "blah"),
            [],
            [],
            id="Complex",
        ),
        pytest.param(
            StructOf(
                mass=ArrayOf(FloatRange()),
                pressure=ArrayOf(FloatRange()),
                timestamp=ArrayOf(FloatRange()),
            ),
            {"mass": [], "pressure": [], "timestamp": []},
            [],
            [],
            id="struct of empty arrays",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [[], []],
            [100, 100],
            [2, 0],
            id="empty 2d array ",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [[]],
            [100, 100],
            [1, 0],
            id="empty 2d array [1,0]",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [],
            [100, 100],
            [0],
            id="empty empty 2d",
        ),
    ],
)
def test_arrayof_update_dtype(start_dtype, data, expected_shape, expected_update_shape):
    sdtype = SECoPdtype(start_dtype)
    assert sdtype.shape == expected_shape

    sdtype.update_dtype(data)
    assert sdtype.shape == expected_update_shape

    SECoPReading(sdtype, CacheItem(data, 234342.2, None, datatype=start_dtype))


@pytest.mark.parametrize(
    "start_dtype,expected_dtype_descr,expected_shape,max_depth",
    [
        pytest.param(
            TupleOf(ArrayOf(FloatRange(), maxlen=50), StringType()),
            [("f0", "<f8", (50,)), ("f1", "<U100")],
            [],
            1,
            id="Tuple of Array",
        ),
        pytest.param(
            ArrayOf(
                TupleOf(FloatRange(), StringType(), ArrayOf(ArrayOf(FloatRange()))),
                maxlen=50,
            ),
            [("f0", "<f8"), ("f1", "<U100"), ("f2", "<f8", (100, 100))],
            [50],
            1,
            id="Array of Tuple",
        ),
        pytest.param(
            ArrayOf(
                TupleOf(
                    FloatRange(), StringType(), TupleOf(FloatRange(), StringType())
                ),
                maxlen=50,
            ),
            [("f0", "<f8"), ("f1", "<U100"), ("f2", [("f0", "<f8"), ("f1", "<U100")])],
            [50],
            2,
            id="Array of (Tuple of Tuple)",
        ),
        pytest.param(
            ArrayOf(
                StructOf(a=FloatRange(), b=StringType(), c=StringType()),
                maxlen=50,
            ),
            [("a", "<f8"), ("b", "<U100"), ("c", "<U100")],
            [50],
            1,
            id="Array of Struct",
        ),
        pytest.param(
            ArrayOf(
                StructOf(
                    a=FloatRange(),
                    b=StringType(),
                    c=TupleOf(FloatRange(), FloatRange()),
                ),
                maxlen=50,
            ),
            [("a", "<f8"), ("b", "<U100"), ("c", [("f0", "<f8"), (("f1", "<f8"))])],
            [50],
            2,
            id="Array of (Struct of tuple)",
        ),
    ],
)
def test_describe_str(start_dtype, expected_dtype_descr, expected_shape, max_depth):
    sdtype = SECoPdtype(start_dtype)

    assert sdtype.shape == expected_shape
    assert sdtype.dtype_descr == expected_dtype_descr
    assert sdtype.max_depth == max_depth


@pytest.mark.parametrize(
    "start_dtype,np_input,expected_output,type_checks,ophy_val",
    [
        pytest.param(
            StructOf(
                float_val=FloatRange(),
                string_val=StringType(),
                int_val=IntRange(),
                scaled_val=ScaledInteger(scale=0.01),
                enum_val=EnumType(ON=1, OFF=0),
            ),
            np.array(
                (3.14, "test", 42, 123, 1),
                dtype=[
                    ("float_val", "<f8"),
                    ("string_val", "<U100"),
                    ("int_val", "<i4"),
                    ("scaled_val", "<f8"),
                    ("enum_val", "<i4"),
                ],
            ),
            {
                "float_val": 3.14,
                "string_val": "test",
                "int_val": 42,
                "scaled_val": 123,
                "enum_val": 1,
            },
            lambda val: (
                isinstance(val["float_val"], float)
                and isinstance(val["string_val"], str)
                and isinstance(val["int_val"], int)
                and isinstance(val["scaled_val"], int)
                and isinstance(val["enum_val"], int)
            ),
            np.array(
                (3.14, "test", 42, 123, 1),
                dtype=[
                    ("float_val", "<f8"),
                    ("string_val", "<U100"),
                    ("int_val", "<i4"),
                    ("scaled_val", "<f8"),
                    ("enum_val", "<i4"),
                ],
            ),
            id="Struct of all atomic types",
        ),
        pytest.param(
            TupleOf(
                FloatRange(),
                StringType(),
                IntRange(),
                ScaledInteger(scale=0.1),
                EnumType(ON=1, OFF=0),
            ),
            np.array(
                (2.718, "hello", 99, 25, 0),
                dtype=[
                    ("f0", "<f8"),
                    ("f1", "<U100"),
                    ("f2", "<i8"),
                    ("f3", "<f8"),
                    ("f4", "<i8"),
                ],
            ),
            (2.718, "hello", 99, 25, 0),
            lambda val: (
                isinstance(val[0], float)
                and isinstance(val[1], str)
                and isinstance(val[2], int)
                and isinstance(val[3], int)
                and isinstance(val[4], int)
            ),
            np.array(
                (2.718, "hello", 99, 25, 0),
                dtype=[
                    ("f0", "<f8"),
                    ("f1", "<U100"),
                    ("f2", "<i8"),
                    ("f3", "<f8"),
                    ("f4", "<i8"),
                ],
            ),
            id="Tuple of all atomic types",
        ),
        pytest.param(
            StructOf(value=FloatRange(), state=EnumType(IDLE=0, BUSY=1, ERROR=2)),
            np.array((42.5, 2), dtype=[("value", "<f8"), ("state", "<i4")]),
            {"value": 42.5, "state": 2},
            lambda val: (
                isinstance(val["value"], float) and isinstance(val["state"], int)
            ),
            np.array((42.5, 2), dtype=[("value", "<f8"), ("state", "<i4")]),
            id="Struct of Float and Enum",
        ),
        pytest.param(
            ArrayOf(EnumType(ON=1, OFF=0)),
            [1, 0, 1],
            ArrayOf(EnumType(ON=1, OFF=0)).import_value([1, 0, 1]),
            lambda val: isinstance(val, tuple)
            and all(isinstance(v, EnumMember) for v in val),
            ["ON", "OFF", "ON"],
            id="Array of Enums",
        ),
        pytest.param(
            ArrayOf(ArrayOf(EnumType(ON=1, OFF=0))),
            [[1, 0, 1], [0], [1]],
            ArrayOf(ArrayOf(EnumType(ON=1, OFF=0))).import_value([[1, 0, 1], [0], [1]]),
            lambda val: isinstance(val, tuple)
            and all(isinstance(v, EnumMember) or isinstance(v, tuple) for v in val),
            [["ON", "OFF", "ON"], ["OFF"], ["ON"]],
            id="Ragged array of Enums",
        ),
    ],
)
def test_val2secop(start_dtype, np_input, expected_output, type_checks, ophy_val):
    sdtype = SECoPdtype(start_dtype)

    secop_val = sdtype.val2secop(np_input)

    # Use numpy.testing for proper comparison with numpy arrays
    np.testing.assert_equal(secop_val, expected_output)

    # Verify types are correct (should be Python types, not numpy arrays)

    assert type_checks(secop_val), f"Type check failed for {secop_val}"

    back_to_ophyd = sdtype.secop2val(secop_val)

    print(f"Original ophyd value: {ophy_val}")
    print(f"Back to ophyd value: {back_to_ophyd}")

    assert (
        back_to_ophyd == ophy_val
    ), f"Back to ophyd conversion failed for {back_to_ophyd}"



import asyncio
import logging
import re
import time as ttime
import warnings
from abc import abstractmethod
from logging import Logger
from types import MethodType
from typing import Any, Dict, Iterator, Optional, Type

import bluesky.plan_stubs as bps
from bluesky.protocols import (
    Flyable,
    Locatable,
    Location,
    PartialEvent,
    Reading,
    Stoppable,
    Subscribable,
    Triggerable,
)
from frappy.datatypes import (
    ArrayOf,
    BLOBType,
    BoolType,
    CommandType,
    FloatRange,
    IntRange,
    ScaledInteger,
    StringType,
    StructOf,
    TupleOf,
)
from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    Device,
    DeviceConnector,
    DeviceFiller,
    LazyMock,
    Signal,
    SignalR,
    SignalRW,
    SignalX,
    StandardReadable,
    StandardReadableFormat,
    observe_value,
)
from ophyd_async.core._utils import Callback

from secop_ophyd.AsyncFrappyClient import AsyncFrappyClient
from secop_ophyd.logs import setup_logging
from secop_ophyd.propertykeys import DATAINFO, EQUIPMENT_ID, INTERFACE_CLASSES
from secop_ophyd.SECoPSignal import (
    AttributeType,
    LocalBackend,
    SECoPBackend,
    SECoPXBackend,
)
from secop_ophyd.util import Path

# Predefined Status Codes
DISABLED = 0
IDLE = 100
STANDBY = 130
PREPARED = 150
WARN = 200
WARN_STANDBY = 230
WARN_PREPARED = 250
NSTABLE = 270  # not in SECoP standard (yet)
BUSY = 300
DISABLING = 310
INITIALIZING = 320
PREPARING = 340
STARTING = 360
RAMPING = 370
STABILIZING = 380
FINALIZING = 390
ERROR = 400
ERROR_STANDBY = 430
ERROR_PREPARED = 450
UNKNOWN = 401  # not in SECoP standard (yet)


IGNORED_PROPS = ["meaning", "_plotly"]


def clean_identifier(anystring):
    return str(re.sub(r"\W+|^(?=\d)", "_", anystring))


def secop_enum_name_to_python(member_name: str) -> str:
    """Convert SECoP enum member name to Python identifier.

    Examples:
        'Low Energy' -> 'LOW_ENERGY'
        'high-power' -> 'HIGH_POWER'
        'Mode 1' -> 'MODE_1'

    :param member_name: Original SECoP enum member name
    :return: Python-compatible identifier in UPPER_CASE
    """
    # Replace spaces and hyphens with underscores, remove other special chars
    cleaned = re.sub(r"[\s-]+", "_", member_name)
    cleaned = re.sub(r"[^a-zA-Z0-9_]", "", cleaned)
    # Convert to uppercase
    cleaned = cleaned.upper()
    # Ensure it doesn't start with a digit
    if cleaned and cleaned[0].isdigit():
        cleaned = "_" + cleaned
    return cleaned


def format_assigned(device: StandardReadable, signal: SignalR) -> bool:
    if (
        signal.describe in device._describe_funcs
        or signal.describe in device._describe_config_funcs
    ):
        # Standard readable format already assigned
        return True

    return False


def is_read_signal(device: StandardReadable, signal: SignalR | SignalRW) -> bool:
    if signal.describe in device._describe_funcs:
        return True

    return False


def is_config_signal(device: StandardReadable, signal: SignalR | SignalRW) -> bool:
    if signal.describe in device._describe_config_funcs:
        return True

    return False


class ParameterType:
    """Annotation for Parameter Signals, defines the path to the parameter
    in the secclient module dict"""

    def __repr__(self) -> str:
        """Return repr suitable for code generation in annotations."""
        return "ParamT()"

    def __call__(self, parent: Device, child: Device):
        if not isinstance(child, Signal):
            return

        backend = child._connector.backend

        if not isinstance(backend, SECoPBackend):
            return

        backend.attribute_type = AttributeType.PARAMETER
        backend._secclient = parent._client


class PropertyType:
    """Annotation for Module Property Signals, defines the path to the property"""

    def __repr__(self) -> str:
        """Return repr suitable for code generation in annotations."""
        return "PropT()"

    def __call__(self, parent: Device, child: Device):
        if not isinstance(child, Signal):
            return

        backend = child._connector.backend

        if not isinstance(backend, SECoPBackend):
            return

        backend.attribute_type = AttributeType.PROPERTY
        backend._secclient = parent._client


class SECoPDeviceConnector(DeviceConnector):

    sri: str
    module: str | None
    node_id: str
    _auto_fill_signals: bool

    def __init__(
        self,
        sri: str,
        auto_fill_signals: bool = True,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ) -> None:

        self.sri = sri
        self.node_id = sri.split(":")[0] + ":" + sri.split(":")[1]
        self._auto_fill_signals = auto_fill_signals
        self.loglevel = loglevel
        self.logdir = logdir

        if sri.count(":") == 2:
            self.module = sri.split(":")[2]
        elif sri.count(":") == 1:
            self.module = None
        else:
            raise RuntimeError(f"Invalid SECoP resource identifier: {sri}")

        if SECoPDevice._clients.get(self.node_id) is None:
            raise RuntimeError(f"No AsyncFrappyClient for URI {sri} exists")

        self.client: AsyncFrappyClient = SECoPDevice._clients[self.node_id]

    def set_module(self, module_name: str):
        if self.sri.count(":") != 1:
            raise RuntimeError(
                "Module can only be set if SRI does not already contain module"
            )
        self.module = module_name
        self.sri = self.sri + ":" + module_name

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            self.filler = DeviceFiller(
                device=device,
                signal_backend_factory=SECoPBackend,
                device_connector_factory=lambda: SECoPDeviceConnector(
                    self.sri, self._auto_fill_signals, self.loglevel, self.logdir
                ),
            )

        list(self.filler.create_signals_from_annotations())
        list(self.filler.create_devices_from_annotations(filled=False))

        self.filler.check_created()

    def fill_backend_with_path(self, backend: SECoPBackend, annotations: list[Any]):
        unhandled = []
        while annotations:
            annotation = annotations.pop(0)

            if isinstance(annotation, StandardReadableFormat):
                backend.format = annotation

            else:
                unhandled.append(annotation)

        annotations.extend(unhandled)
        # These leftover annotations will now be handled by the iterator

    async def connect_mock(self, device: Device, mock: LazyMock):
        # Make 2 entries for each DeviceVector
        self.filler.create_device_vector_entries_to_mock(2)
        # Set the name of the device to name all children
        device.set_name(device.name)
        await super().connect_mock(device, mock)

    async def connect_real(self, device: Device, timeout: float, force_reconnect: bool):
        if not self.sri:
            raise RuntimeError(f"Could not connect to SEC node: {self.sri}")

        # Establish connection to SEC Node
        await self.client.connect(3)

        # Module Device: fill Parameters & Pproperties
        # (commands are done via annotated plans)
        if self.module:

            # Fill Parmeters
            parameter_dict = self.client.modules[self.module]["parameters"]
            # remove ignored signals
            parameters = [
                child
                for child in parameter_dict.keys()
                if child not in self.filler.ignored_signals
            ]

            # Dertermine children that are declared but not yet filled
            not_filled = {unfilled for unfilled, _ in device.children()}

            for param_name in parameters:
                if self._auto_fill_signals or param_name in not_filled:
                    signal_type = (
                        SignalR if parameter_dict[param_name]["readonly"] else SignalRW
                    )

                    backend = self.filler.fill_child_signal(param_name, signal_type)

                    from secop_ophyd.GenNodeCode import get_type_param

                    datatype = get_type_param(parameter_dict[param_name]["datatype"])
                    backend.init_parameter_from_introspection(
                        datatype=datatype,
                        path=self.module + ":" + param_name,
                        secclient=self.client,
                    )

            # Fill Properties
            module_property_dict = self.client.modules[self.module]["properties"]

            # remove ignored signals
            module_properties = [
                child
                for child in module_property_dict.keys()
                if child not in self.filler.ignored_signals
                and child not in IGNORED_PROPS
            ]

            for mod_property_name in module_properties:
                if self._auto_fill_signals or mod_property_name in not_filled:

                    # properties are always read only
                    backend = self.filler.fill_child_signal(mod_property_name, SignalR)

                    from secop_ophyd.GenNodeCode import get_type_prop

                    datatype = get_type_prop(module_property_dict[mod_property_name])

                    backend.init_property_from_introspection(
                        datatype=datatype,
                        path=self.module + ":" + mod_property_name,
                        secclient=self.client,
                    )

        # Node Device: fill child devices (modules)
        else:

            # Fill Module devices
            modules = self.client.modules

            not_filled = {unfilled for unfilled, _ in device.children()}

            for module_name in modules.keys():
                if self._auto_fill_signals or module_name in not_filled:
                    module_properties = modules[module_name]["properties"]
                    device_sub_class = class_from_interface(module_properties)

                    self.filler.fill_child_device(module_name, device_sub_class)

                    mod_dev: SECoPDevice = getattr(device, module_name)
                    mod_dev.set_module(module_name)

            # Fill Node properties
            node_property_dict = self.client.properties

            # remove ignored signals
            node_properties = [
                child
                for child in node_property_dict.keys()
                if child not in self.filler.ignored_signals
            ]

            for node_property_name in node_properties:
                if self._auto_fill_signals or node_property_name in not_filled:

                    # properties are always read only
                    backend = self.filler.fill_child_signal(node_property_name, SignalR)

                    from secop_ophyd.GenNodeCode import get_type_prop

                    datatype = get_type_prop(node_property_dict[node_property_name])

                    backend.init_property_from_introspection(
                        datatype=datatype,
                        path=node_property_name,
                        secclient=self.client,
                    )

        self.filler.check_filled(f"{self.node_id}")

        # Set the name of the device to name all children
        device.set_name(device.name)
        await super().connect_real(device, timeout, force_reconnect)

        # All Signals and child devs should be filled and connected now, in the next
        # all signals and child devices need to be added to the according
        # StandardReadableFormat with the hierarchiy:
        # 1. Format given in Annotation
        #       --> these will already have been set by the DeviceFiller
        # 2. Module Interface Class definition (value, target,...)
        #       --> these are set at the end of a .connect() method of the according
        #           SECoPDevice subclass skipping any signals that have already been
        #           set by annotations (should emit warning if there is a conflict
        #           config vs read sig)
        # 3. & 4. Definition in Parameter property "_signal_format" + Defaults
        #   - _signal_format property + default CONFIG_SIGNAL for all other Signals
        #       --> these are set here at the end of SECoPDeviceConnector.connect_real()
        #           for all Signals that have not yet been set to a format
        #   - CHILD format for all child devices (SECoPDevice instances)
        #       --> these are set at the end of SECoPNodeDevice.connect() method
        #           the device tree has only a depth of 2 levels (Node -> Modules)
        #

        # device has to be standard readable for this to make sense
        if not isinstance(device, SECoPDevice):
            return

        # 2. Module Interface Class definition (value, target,...)
        await device._assign_interface_formats()

        # 3. & 4. Definition in Parameter property "_signal_format" + Defaults
        await device._assign_default_formats()


class SECoPCMDDevice(StandardReadable, Flyable, Triggerable):
    """
    Command devices that have Signals for command args, return values and a signal
    for triggering command execution (SignalX). They themselves are triggerable.

    Once the CMD Device is triggered, the command args are retrieved from the 'argument'
    Signal. The command message is sent to the SEC Node and the return value is written
    to 'result' signal.

    """

    def __init__(self, path: Path, secclient: AsyncFrappyClient):
        """Initialize the CMD Device

        :param path: Path to the command in the secclient module dict
        :type path: Path
        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        """
        dev_name: str = path.get_signal_name() + "_CMD"

        self._secclient: AsyncFrappyClient = secclient

        cmd_props = secclient.modules[path._module_name]["commands"][
            path._accessible_name
        ]  # noqa: E501
        cmd_datatype: CommandType = cmd_props["datatype"]
        datainfo = cmd_props[DATAINFO]

        self.description: str = cmd_props["description"]
        self.arg_dtype = cmd_datatype.argument
        self.res_dtype = cmd_datatype.result

        self.argument: SignalRW | None
        self.result: SignalR | None

        # result signals
        read = []
        # argument signals
        config = []

        self._start_time: float
        self.commandx: SignalX

        self.wait_idle: bool = False

        with self.add_children_as_readables(
            format=StandardReadableFormat.CONFIG_SIGNAL
        ):
            # Argument Signals (config Signals, can also be read)
            arg_path = path.append("argument")
            if self.arg_dtype is None:
                self.argument = None
            else:
                arg_backend = LocalBackend(
                    path=arg_path,
                    secop_dtype_obj=self.arg_dtype,
                    sig_datainfo=datainfo["argument"],
                )
                self.argument = SignalRW(arg_backend)
                config.append(self.argument)

            # Result Signals  (read Signals)
            res_path = path.append("result")

            if self.res_dtype is None:
                self.result = None
            else:
                res_backend = LocalBackend(
                    path=res_path,
                    secop_dtype_obj=self.res_dtype,
                    sig_datainfo=datainfo["result"],
                )
                self.result = SignalRW(res_backend)
                read.append(self.argument)

            argument = None
            result = None
            if isinstance(self.argument, SignalR):
                argument = self.argument._connector.backend

            if isinstance(self.result, SignalR):
                result = self.result._connector.backend

            # SignalX (signal that triggers execution of the Command)
            exec_backend = SECoPXBackend(
                path=path,
                secclient=secclient,
                argument=argument,  # type: ignore
                result=result,  # type: ignore
            )

        self.commandx = SignalX(exec_backend)

        super().__init__(name=dev_name)

    def trigger(self) -> AsyncStatus:
        """Triggers the SECoPCMDDevice and sends command message to SEC Node.
        Command argument is taken form 'argument' Signal, and return value is
        written in the 'return' Signal

        :return: A Status object, that is marked Done once the answer from the
        SEC Node is received
        :rtype: AsyncStatus
        """
        coro = asyncio.wait_for(fut=self._exec_cmd(), timeout=None)
        return AsyncStatus(awaitable=coro)

    def kickoff(self) -> AsyncStatus:
        # trigger execution of secop command, wait until Device is Busy

        self._start_time = ttime.time()
        coro = asyncio.wait_for(fut=asyncio.sleep(1), timeout=None)
        return AsyncStatus(coro)

    async def _exec_cmd(self):
        stat = self.commandx.trigger()

        await stat

    def complete(self) -> AsyncStatus:
        coro = asyncio.wait_for(fut=self._exec_cmd(), timeout=None)
        return AsyncStatus(awaitable=coro)

    def collect(self) -> Iterator[PartialEvent]:
        yield dict(
            time=self._start_time, timestamps={self.name: []}, data={self.name: []}
        )


class SECoPDevice(StandardReadable):

    _clients: Dict[str, AsyncFrappyClient] = {}

    _node_id: str
    _sri: str
    _host: str
    _port: str
    _module: str | None
    _mod_prop_devices: Dict[str, SignalR]
    _param_devices: Dict[str, Any]
    _logger: Logger

    hinted_signals: list[str] = []

    def __init__(
        self,
        sri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        connector: SECoPDeviceConnector | None = None,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ) -> None:

        if connector and sri:
            raise RuntimeError("Provide either sri or connector, not both")

        if connector:
            sri = connector.sri
            loglevel = connector.loglevel
            logdir = connector.logdir

        self._sri = sri
        self._host = sri.split(":")[0]
        self._port = sri.split(":")[1]
        self._mod_prop_devices = {}
        self._param_devices = {}
        self._node_id = sri.split(":")[0] + ":" + sri.split(":")[1]

        self._logger = setup_logging(
            name=f"frappy:{self._host}:{self._port}",
            level=loglevel,
            log_dir=logdir,
        )

        self._module = None
        if len(sri.split(":")) > 2:
            self._module = sri.split(":")[2]

        if SECoPDevice._clients.get(self._node_id) is None:
            SECoPDevice._clients[self._node_id] = AsyncFrappyClient(
                host=self._host, port=self._port, log=self._logger
            )

        connector = connector or SECoPDeviceConnector(sri=sri)

        self._client: AsyncFrappyClient = SECoPDevice._clients[self._node_id]

        super().__init__(name=name, connector=connector)

    def set_module(self, module_name: str):
        if self._module is not None:
            raise RuntimeError("Module can only be set if it was not already set")

        self._module = module_name
        self._sri = self._sri + ":" + module_name

        self._connector.set_module(module_name)

    async def connect(
        self,
        mock: bool | LazyMock = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ):
        if not self._client.online or force_reconnect:
            # Establish connection to SEC Node
            await self._client.connect(3)

        if self._module:
            module_desc = self._client.modules[self._module]

            # Initialize Command Devices
            for command, _ in module_desc["commands"].items():
                # generate new root path
                cmd_path = Path(parameter_name=command, module_name=self._module)
                cmd_dev_name = command + "_CMD"
                setattr(
                    self,
                    cmd_dev_name,
                    SECoPCMDDevice(path=cmd_path, secclient=self._client),
                )

                cmd_dev: SECoPCMDDevice = getattr(self, cmd_dev_name)
                # Add Bluesky Plan Methods

                # Stop is already an ophyd native operation
                if command == "stop":
                    continue

                cmd_plan = self.generate_cmd_plan(
                    cmd_dev, cmd_dev.arg_dtype, cmd_dev.res_dtype
                )

                setattr(self, command, MethodType(cmd_plan, self))

        await super().connect(mock, timeout, force_reconnect)

        if self._module is None:
            # set device name from equipment id property
            self.set_name(self._client.properties[EQUIPMENT_ID].replace(".", "-"))
        else:
            self.set_name(self._module)

    def generate_cmd_plan(
        self,
        cmd_dev: SECoPCMDDevice,
        argument_type: Type | None = None,
        return_type: Type | None = None,
    ):

        def command_plan_no_arg(self, wait_for_idle: bool = False):
            # Trigger the Command device, meaning that the command gets sent to the
            # SEC Node
            yield from bps.trigger(cmd_dev, wait=True)

            if wait_for_idle:

                def wait_for_idle_factory():
                    return self.wait_for_idle()

                yield from bps.wait_for([wait_for_idle_factory])

            if (
                return_type is not None
                and isinstance(cmd_dev.result, SignalR)
                and isinstance(cmd_dev.result._connector.backend, LocalBackend)
            ):

                return cmd_dev.result._connector.backend.reading.get_value()

        def command_plan(self, arg, wait_for_idle: bool = False):
            # TODO  Type checking

            if arg is not None:
                yield from bps.abs_set(cmd_dev.argument, arg)

            # Trigger the Command device, meaning that the command gets sent to the
            # SEC Node
            yield from bps.trigger(cmd_dev, wait=True)

            if wait_for_idle:

                def wait_for_idle_factory():
                    return self.wait_for_idle()

                yield from bps.wait_for([wait_for_idle_factory])

            if (
                return_type is not None
                and isinstance(cmd_dev.result, SignalR)
                and isinstance(cmd_dev.result._connector.backend, LocalBackend)
            ):

                return cmd_dev.result._connector.backend.reading.get_value()

        cmd_meth = command_plan_no_arg if argument_type is None else command_plan

        anno_dict = cmd_meth.__annotations__

        dtype_mapping = {
            StructOf: dict[str, Any],
            ArrayOf: list[Any],
            TupleOf: tuple[Any],
            BLOBType: str,
            BoolType: bool,
            FloatRange: float,
            IntRange: int,
            ScaledInteger: int,
            StringType: str,
        }

        if return_type is not None:
            anno_dict["return"] = dtype_mapping[return_type.__class__]
        if argument_type is not None:
            anno_dict["arg"] = dtype_mapping[argument_type.__class__]

        return cmd_meth

    @abstractmethod
    async def _assign_interface_formats(self):
        """Assign signal formats specific to this device's interface class.
        Subclasses override this to assign formats before default fallback."""

    async def _assign_default_formats(self):
        config_signals = []
        hinted_signals = []
        uncached_signals = []
        hinted_uncached_signals = []

        def assert_device_is_signalr(device: Device) -> SignalR:
            if not isinstance(device, SignalR):
                raise TypeError(f"{device} is not a SignalR")
            return device

        for _, child in self.children():

            if not isinstance(child, Signal):
                continue

            backend = child._connector.backend

            if not isinstance(backend, SECoPBackend):
                continue

            param_name = backend.path_str.split(":")[-1]
            if param_name == "status":
                # status signals should not be assigned a format,
                # but a SignalR children (this can be removed once tiled can
                # hanlde composite dtypes)
                continue

            # child is a Signal with SECoPParamBackend

            # check if signal already has a format assigned
            signalr_device = assert_device_is_signalr(child)

            if format_assigned(self, signalr_device):
                # format already assigned by annotation or module IF class
                continue

            match backend.format:
                case StandardReadableFormat.CHILD:
                    raise RuntimeError("Signal cannot have CHILD format")
                case StandardReadableFormat.CONFIG_SIGNAL:
                    config_signals.append(signalr_device)
                case StandardReadableFormat.HINTED_SIGNAL:
                    hinted_signals.append(signalr_device)
                case StandardReadableFormat.UNCACHED_SIGNAL:
                    uncached_signals.append(signalr_device)
                case StandardReadableFormat.HINTED_UNCACHED_SIGNAL:
                    hinted_uncached_signals.append(signalr_device)

        # add signals to device in the order of their priority
        self.add_readables(config_signals, StandardReadableFormat.CONFIG_SIGNAL)

        self.add_readables(hinted_signals, StandardReadableFormat.HINTED_SIGNAL)

        self.add_readables(uncached_signals, StandardReadableFormat.UNCACHED_SIGNAL)

        self.add_readables(
            hinted_uncached_signals, StandardReadableFormat.HINTED_UNCACHED_SIGNAL
        )


class SECoPNodeDevice(SECoPDevice):

    hinted_signals: list[str] = []

    def __init__(
        self,
        sec_node_uri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        loglevel=logging.INFO,
        logdir: str | None = None,
    ):
        # ensure sec_node_uri only contains host:port
        if sec_node_uri.count(":") != 1:
            raise RuntimeError(
                f"SECoPNodeDevice SRI must only contain host:port {sec_node_uri}"
            )

        super().__init__(sri=sec_node_uri, name=name, loglevel=loglevel, logdir=logdir)

    async def connect(self, mock=False, timeout=DEFAULT_TIMEOUT, force_reconnect=False):
        await super().connect(mock, timeout, force_reconnect)

        moddevs = []
        for _, moddev in self.children():
            if isinstance(moddev, SECoPDevice):
                moddevs.append(moddev)

        self.add_readables(moddevs, StandardReadableFormat.CHILD)

        # register secclient callbacks (these are useful if sec node description
        # changes after a reconnect)
        self._client.register_callback(
            None, self.descriptiveDataChange, self.nodeStateChange
        )

    def descriptiveDataChange(self, module, description):  # noqa: N802
        raise RuntimeError(
            "The descriptive data has changed upon reconnect. Descriptive data changes"
            "are not supported: reinstantiate device"
        )

    def nodeStateChange(self, online, state):  # noqa: N802
        """called when the state of the connection changes

        'online' is True when connected or reconnecting, False when disconnected
        or connecting 'state' is the connection state as a string
        """
        if state == "connected" and online is True:
            self._client.conn_timestamp = ttime.time()

    async def _assign_interface_formats(self):
        # Node device has no specific interface class formats
        pass

    def class_from_instance(self, path_to_module: str | None = None):
        from secop_ophyd.GenNodeCode import GenNodeCode

        description = self._client.client.request("describe")[2]

        # parse genClass file if already present
        genCode = GenNodeCode(path=path_to_module, log=self._logger)

        genCode.from_json_describe(description)

        genCode.write_gen_node_class_file()


class SECoPCommunicatorDevice(SECoPDevice):

    hinted_signals: list[str] = []

    def __init__(
        self,
        sri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        connector: SECoPDeviceConnector | None = None,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ) -> None:
        super().__init__(
            sri=sri, name=name, connector=connector, loglevel=loglevel, logdir=logdir
        )

    async def _assign_interface_formats(self):
        # Communicator has no specific interface class formats
        pass


class SECoPReadableDevice(SECoPDevice, Triggerable, Subscribable):
    """
    Standard readable SECoP device, corresponding to a SECoP module with the
    interface class "Readable"
    """

    hinted_signals: list[str] = ["value"]

    def __init__(
        self,
        sri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        connector: SECoPDeviceConnector | None = None,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ):
        """Initializes the SECoPReadableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: Name of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.value: SignalR
        self.status: SignalR

        super().__init__(
            sri=sri, name=name, connector=connector, loglevel=loglevel, logdir=logdir
        )

    async def connect(self, mock=False, timeout=DEFAULT_TIMEOUT, force_reconnect=False):
        await super().connect(mock, timeout, force_reconnect)

        if not hasattr(self, "value"):
            raise AttributeError(
                "Attribute 'value' has not been assigned,"
                + "but is needed for Readable interface class"
            )

        if not hasattr(self, "status"):
            raise AttributeError(
                "Attribute 'status' has not been assigned,"
                + "but is needed for Readable interface class"
            )

    async def _assign_interface_formats(self):

        if format_assigned(self, self.value):
            if not is_read_signal(self, self.value):
                warnings.warn(
                    f"Signal 'value' of device {self.name} has format assigned "
                    + "that is not compatible with Readable interface class"
                )
        else:
            self.add_readables([self.value], StandardReadableFormat.HINTED_SIGNAL)

        # TODO ensure status signal must be neither config nor read format

    async def wait_for_idle(self):
        """asynchronously waits until module is IDLE again. this is helpful,
        for running commands that are not done immediately
        """

        self._logger.info(f"Waiting for {self.name} to be IDLE")

        if self.status is None:
            self._logger.error("Status Signal not initialized")
            raise Exception("status Signal not initialized")

        # force reading of fresh status from device
        await self.status.read(False)

        async for current_stat in observe_value(self.status):
            # status is has type Tuple and is therefore transported as
            # structured Numpy array ('f0':statuscode;'f1':status Message)

            stat_code = current_stat["f0"]

            # Module is in IDLE/WARN state
            if IDLE <= stat_code < BUSY:
                self._logger.info(f"Module {self.name} --> IDLE")
                break

            if hasattr(self, "_stopped"):
                # self.logger.info(f"Module {self.name} was stopped STOPPED")
                if self._stopped is True:
                    break

            # Error State or DISABLED
            if hasattr(self, "_success"):
                if stat_code >= ERROR or stat_code < IDLE:
                    self._logger.error(f"Module {self.name} --> ERROR/DISABLED")
                    self._success = False
                    break

    # TODO add timeout
    def observe_status_change(self, monitored_status_code: int):
        async def switch_from_status_inner():
            async for current_stat in observe_value(self.status):
                # status is has type Tuple and is therefore transported as
                # structured Numpy array ('f0':statuscode;'f1':status Message)

                stat_code = current_stat["f0"]

                if monitored_status_code != stat_code:
                    break

        def switch_from_status_factory():
            return switch_from_status_inner()

        yield from bps.wait_for([switch_from_status_factory])

    def trigger(self) -> AsyncStatus:
        self._logger.info(f"Triggering {self.name}: read fresh data from device")
        # get fresh reading of the value Parameter from the SEC Node
        return AsyncStatus(
            awaitable=self._client.get_parameter(self._module, "value", trycache=False)
        )

    def subscribe(self, function: Callback[dict[str, Reading]]) -> None:
        """Subscribe to updates in the reading"""
        self.value.subscribe(function=function)

    def clear_sub(self, function: Callback) -> None:
        """Remove a subscription."""
        self.value.clear_sub(function=function)


class SECoPTriggerableDevice(SECoPReadableDevice, Stoppable):
    """
    Standard triggerable SECoP device, corresponding to a SECoP module with the0s
    interface class "Triggerable"
    """

    hinted_signals: list[str] = ["value"]

    def __init__(
        self,
        sri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        connector: SECoPDeviceConnector | None = None,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ):
        """Initialize SECoPTriggerableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: ame of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.go_CMD: SECoPCMDDevice

        self._success = True
        self._stopped = False

        super().__init__(
            sri=sri, name=name, connector=connector, loglevel=loglevel, logdir=logdir
        )


class SECoPWritableDevice(SECoPReadableDevice):
    hinted_signals: list[str] = ["target", "value"]

    pass


class SECoPMoveableDevice(SECoPReadableDevice, Locatable, Stoppable):
    """
    Standard movable SECoP device, corresponding to a SECoP module with the
    interface class "Drivable"
    """

    hinted_signals: list[str] = ["target", "value"]

    def __init__(
        self,
        sri: str = "",  # SECoP resource identifier host:port:optional[module]
        name: str = "",
        connector: SECoPDeviceConnector | None = None,
        loglevel=logging.INFO,
        logdir: str | None = None,
    ):
        """Initialize SECoPMovableDevice

        :param secclient: SECoP client providing communication to the SEC Node
        :type secclient: AsyncFrappyClient
        :param module_name: ame of the SEC Node module that is represented by
            this device
        :type module_name: str
        """

        self.target: SignalRW

        super().__init__(
            sri=sri, name=name, connector=connector, loglevel=loglevel, logdir=logdir
        )

        self._success = True
        self._stopped = False

    async def connect(self, mock=False, timeout=DEFAULT_TIMEOUT, force_reconnect=False):

        await super().connect(mock, timeout, force_reconnect)

        if not hasattr(self, "target"):
            raise AttributeError(
                "Attribute 'target' has not been assigned, "
                + "but is needed for 'Drivable' interface class!"
            )

    def set(self, new_target, timeout: Optional[float] = None) -> AsyncStatus:
        """Sends new target to SEC Nonde and waits until module is IDLE again

        :param new_target: new taget/setpoint for module
        :type new_target: _type_
        :param timeout: timeout for set operation, defaults to None
        :type timeout: Optional[float], optional
        :return: Asyncstatus that gets set to Done once module is IDLE again
        :rtype: AsyncStatus
        """
        coro = asyncio.wait_for(self._move(new_target), timeout=timeout)
        return AsyncStatus(coro)

    async def _move(self, new_target):
        self._success = True
        self._stopped = False

        await self.target.set(new_target)
        self._logger.info(f"Moving {self.name} to {new_target}")

        # force reading of status from device
        await self.status.read(False)

        # observe status and wait until dvice is IDLE again
        async for current_stat in observe_value(self.status):
            stat_code = current_stat["f0"]

            if self._stopped is True:
                self._logger.info(
                    f"Move of {self.name} to {new_target} was stopped STOPPED"
                )
                break

            # Error State or DISABLED
            if stat_code >= ERROR or stat_code < IDLE:
                self._logger.error(f"Module {self.name} --> ERROR/DISABLED")
                self._success = False
                break

            # Module is in IDLE/WARN state
            if IDLE <= stat_code < BUSY:
                self._logger.info(f"Reached Target Module {self.name} --> IDLE")
                break

            # TODO other status transitions

        if not self._success:
            self._logger.error(
                f"Move of {self.name} to {new_target} was not successful"
            )

    async def stop(self, success=True):
        """Calls stop command on the SEC Node module

        :param success:
            True: device is stopped as planned
            False: something has gone wrong
            (defaults to True)
        :type success: bool, optional
        """
        self._success = success

        if not success:
            self._logger.info(f"Stopping {self.name} success={success}")
            await self._client.exec_command(self._module, "stop")
            self._stopped = True

    async def locate(self) -> Location:
        # return current location of the device (setpoint and readback).
        # Only locally cached values are returned
        setpoint = await self._client.get_parameter(self._module, "target", True)
        readback = await self._client.get_parameter(self._module, "value", True)

        location: Location = {
            "setpoint": setpoint.value,
            "readback": readback.value,
        }
        return location

    async def _assign_interface_formats(self):
        await super()._assign_interface_formats()

        if format_assigned(self, self.target):
            if not is_read_signal(self, self.target):
                warnings.warn(
                    f"Signal 'target' of device {self.name} has format assigned "
                    + "that is not compatible with Movable interface class"
                )
        else:
            self.add_readables([self.target], StandardReadableFormat.HINTED_SIGNAL)


def class_from_interface(mod_properties: dict):
    ophyd_class = None

    # infer highest level IF class
    module_interface_classes: dict = mod_properties[INTERFACE_CLASSES]
    for interface_class in IF_CLASSES.keys():
        if interface_class in module_interface_classes:
            ophyd_class = IF_CLASSES[interface_class]
            break

    # No predefined IF class was a match --> use base class (loose collection of
    # accessibles)
    if ophyd_class is None:
        ophyd_class = SECoPDevice  # type: ignore

    return ophyd_class


IF_CLASSES = {
    "Triggerable": SECoPTriggerableDevice,
    "Drivable": SECoPMoveableDevice,
    "Writable": SECoPWritableDevice,
    "Readable": SECoPReadableDevice,
    "Communicator": SECoPCommunicatorDevice,
}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import secop_ophyd

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "SECoP-Ophyd"
copyright = "2024, Peter Braun"
author = "Peter Braun"
release = secop_ophyd.__version__

# Clean up version for display - extract just X.Y.Z from version strings like "0.13.9.dev0+gb43270d29.d20251029"
if "+" in secop_ophyd.__version__:
    # Development version - use short version or "dev"
    release = secop_ophyd.__version__.split("+")[0]
else:
    release = secop_ophyd.__version__

version = secop_ophyd.__version__

language = "en"

source_suffix = ".rst"
master_doc = "index"


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # for diagrams
    "sphinxcontrib.mermaid",
    # Use this for generating API docs
    "autodoc2",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
    # Add a copy button to each code block
    "sphinx_copybutton",
    # For the card element
    "sphinx_design",
    # To make .nojekyll
    "sphinx.ext.githubpages",
    # To make the {ipython} directive
    "IPython.sphinxext.ipython_directive",
    # To syntax highlight "ipython" language code blocks
    "IPython.sphinxext.ipython_console_highlighting",
    # To embed matplotlib plots generated from code
    "matplotlib.sphinxext.plot_directive",
    # To parse markdown
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = []  # type: ignore


# Which package to load and document
autodoc2_packages = [{"path": "../../src/secop_ophyd", "auto_mode": True}]

# Put them in docs/_api which is git ignored
autodoc2_output_dir = "_api"


# Don't document private things
autodoc2_hidden_objects = {"private", "dunder", "inherited"}

# MyST parser extensions
myst_enable_extensions = ["colon_fence", "fieldlist"]

# Intersphinx configuration for linking to external documentation
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "bluesky": ("https://blueskyproject.io/bluesky/main", None),
    "ophyd-async": ("https://blueskyproject.io/ophyd-async/main", None),
    "numpy": ("https://numpy.org/devdocs/", None),
}

# Set copy-button to ignore python and bash prompts
copybutton_prompt_text = (
    r">>> |\\.\\.\\. |\\$ |In \\[\\d*\\]: | {2,5}\\.\\.\\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# Generate the API documentation when building
autosummary_generate = True
numpydoc_show_class_members = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options for pydata_sphinx_theme
html_theme_options = {
    "use_edit_page_button": True,
    "github_url": "https://github.com/SampleEnvironment/secop-ophyd",
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/secop-ophyd",
            "icon": "fas fa-cube",
        },
    ],
    "external_links": [
        {
            "name": "SECoP Specification",
            "url": "https://sampleenvironment.github.io/secop-site/",
        },
    ],
    "navigation_with_keys": False,
    "show_toc_level": 3,
}

html_logo = "../images/logo.svg"
html_favicon = "../images/favicon.ico"

html_context = {
    "github_user": "SampleEnvironment",
    "github_repo": "secop-ophyd",
    "github_version": "main",
    "doc_path": "docs",
}

# If true, "Created using Sphinx" is shown in the HTML footer.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer.
html_show_copyright = False

# Custom CSS
html_css_files = ["custom.css"]



"""Test full code generation with enums."""

from secop_ophyd.GenNodeCode import (
    EnumClass,
    EnumMember,
    GenNodeCode,
    ModuleClass,
    ParameterAttribute,
)


def test_code_generation_with_enums(clean_generated_file):
    """Test that generated code includes correct enum base classes."""

    gen_code = GenNodeCode(path=clean_generated_file)

    # Create two module classes with different enums for the same parameter
    enum1 = EnumClass(
        name="MassflowControllerGastypeEnum",
        members=[
            EnumMember("H2", "H2", "Hydrogen"),
            EnumMember("N2", "N2", "Nitrogen"),
        ],
        description="Gastype enum for MassflowController",
    )

    enum2 = EnumClass(
        name="MassflowControllerGastypeEnum",
        members=[
            EnumMember("N2", "N2", "Nitrogen"),
            EnumMember("AR", "Ar", "Argon"),
        ],
        description="Gastype enum for MassflowController",
    )

    mod1 = ModuleClass(
        name="MassflowController1",
        bases=["Device"],
        parameters=[
            ParameterAttribute("gastype", "SignalRW", "MassflowControllerGastypeEnum"),
        ],
        properties=[],
        enums=[enum1],
    )

    mod2 = ModuleClass(
        name="MassflowController2",
        bases=["Device"],
        parameters=[
            ParameterAttribute("gastype", "SignalRW", "MassflowControllerGastypeEnum"),
        ],
        properties=[],
        enums=[enum2],
    )

    gen_code.module_classes = [mod1, mod2]

    # Generate code
    code = gen_code.generate_code()

    print("Generated code:")
    print("=" * 60)
    print(code)
    print("=" * 60)

    # Verify the enum uses SupersetEnum
    assert "class MassflowControllerGastypeEnum(SupersetEnum):" in code
    assert "SupersetEnum" in code

    # Verify all three members are present
    assert 'H2 = "H2"' in code
    assert 'N2 = "N2"' in code
    assert 'AR = "Ar"' in code

    print("\n✓ Code generation with SupersetEnum successful!")


def test_code_generation_strict_enum(clean_generated_file):
    """Test that identical enums generate StrictEnum."""

    gen_code = GenNodeCode(path=clean_generated_file)

    # Create two module classes with identical enums
    enum1 = EnumClass(
        name="StatusEnum",
        members=[
            EnumMember("IDLE", "idle", "Idle state"),
            EnumMember("BUSY", "busy", "Busy state"),
        ],
        description="Status enum",
    )

    enum2 = EnumClass(
        name="StatusEnum",
        members=[
            EnumMember("IDLE", "idle", "Idle state"),
            EnumMember("BUSY", "busy", "Busy state"),
        ],
        description="Status enum",
    )

    mod1 = ModuleClass(
        name="Device1",
        bases=["Device"],
        parameters=[
            ParameterAttribute("status", "SignalR", "StatusEnum"),
        ],
        properties=[],
        enums=[enum1],
    )

    mod2 = ModuleClass(
        name="Device2",
        bases=["Device"],
        parameters=[
            ParameterAttribute("status", "SignalR", "StatusEnum"),
        ],
        properties=[],
        enums=[enum2],
    )

    gen_code.module_classes = [mod1, mod2]

    # Generate code
    code = gen_code.generate_code()

    print("\nGenerated code with StrictEnum:")
    print("=" * 60)
    print(code)
    print("=" * 60)

    # Verify the enum uses StrictEnum
    assert "class StatusEnum(StrictEnum):" in code

    print("\n✓ Code generation with StrictEnum successful!")



"""Test enum merging logic for StrictEnum vs SupersetEnum."""

from secop_ophyd.GenNodeCode import (
    EnumClass,
    EnumMember,
    GenNodeCode,
    ModuleClass,
    ParameterAttribute,
)


def test_identical_enums_use_strict(clean_generated_file):
    """When module instances have identical enum values, use StrictEnum."""
    gen_code = GenNodeCode(clean_generated_file)

    # Create two module classes with identical enums
    enum1 = EnumClass(
        name="TestParamEnum",
        members=[
            EnumMember("H2", "H2", "Hydrogen"),
            EnumMember("N2", "N2", "Nitrogen"),
        ],
    )

    enum2 = EnumClass(
        name="TestParamEnum",
        members=[
            EnumMember("H2", "H2", "Hydrogen"),
            EnumMember("N2", "N2", "Nitrogen"),
        ],
    )

    # Add module classes with these enums
    gen_code.module_classes = []

    mod1 = ModuleClass(name="Module1", bases=["Device"], enums=[enum1])
    mod2 = ModuleClass(name="Module2", bases=["Device"], enums=[enum2])

    gen_code.module_classes = [mod1, mod2]

    # Collect enums
    merged = gen_code._collect_all_enums()

    # Should have one enum using StrictEnum
    assert len(merged) == 1
    assert merged[0].base_enum_class == "StrictEnum"
    assert len(merged[0].members) == 2
    print("✓ Identical enums correctly use StrictEnum")


def test_different_enums_use_superset(clean_generated_file):
    """When module instances have different enum values, use SupersetEnum."""
    gen_code = GenNodeCode(clean_generated_file)

    # Create two module classes with different enum members
    enum1 = EnumClass(
        name="TestParamEnum",
        members=[
            EnumMember("H2", "H2", "Hydrogen"),
            EnumMember("N2", "N2", "Nitrogen"),
        ],
    )

    enum2 = EnumClass(
        name="TestParamEnum",
        members=[
            EnumMember("N2", "N2", "Nitrogen"),
            EnumMember("AR", "Ar", "Argon"),
        ],
    )

    # Add module classes with these enums
    from secop_ophyd.GenNodeCode import ModuleClass

    mod1 = ModuleClass(name="Module1", bases=["Device"], enums=[enum1])
    mod2 = ModuleClass(name="Module2", bases=["Device"], enums=[enum2])

    gen_code.module_classes = [mod1, mod2]

    # Collect enums
    merged = gen_code._collect_all_enums()

    # Should have one enum using SupersetEnum with all members
    assert len(merged) == 1
    assert merged[0].base_enum_class == "SupersetEnum"
    assert len(merged[0].members) == 3  # H2, N2, AR

    member_names = {m.name for m in merged[0].members}
    assert member_names == {"H2", "N2", "AR"}
    print("✓ Different enums correctly merged into SupersetEnum")


def test_single_enum_uses_strict(clean_generated_file):
    """When only one module has an enum, use StrictEnum."""
    gen_code = GenNodeCode(clean_generated_file)

    enum1 = EnumClass(
        name="TestParamEnum",
        members=[
            EnumMember("H2", "H2", "Hydrogen"),
            EnumMember("N2", "N2", "Nitrogen"),
        ],
    )

    from secop_ophyd.GenNodeCode import ModuleClass

    mod1 = ModuleClass(name="Module1", bases=["Device"], enums=[enum1])

    gen_code.module_classes = [mod1]

    # Collect enums
    merged = gen_code._collect_all_enums()

    # Should have one enum using StrictEnum
    assert len(merged) == 1
    assert merged[0].base_enum_class == "StrictEnum"
    assert len(merged[0].members) == 2
    print("✓ Single enum correctly uses StrictEnum")


def test_same_class_different_instances(clean_generated_file):
    """When same module class has different instances with different enums, use
    SupersetEnum."""
    gen_code = GenNodeCode(clean_generated_file)

    # Simulate adding the same module class twice (different instances)
    # First instance with AR, N2, H2
    enum1 = EnumClass(
        name="Test_EnumGas_typeEnum",
        members=[
            EnumMember("AR", "AR", "Argon"),
            EnumMember("N2", "N2", "Nitrogen"),
            EnumMember("H2", "H2", "Hydrogen"),
        ],
    )

    # Second instance with AR, He, CO2
    enum2 = EnumClass(
        name="Test_EnumGas_typeEnum",
        members=[
            EnumMember("AR", "AR", "Argon"),
            EnumMember("HE", "He", "Helium"),
            EnumMember("CO2", "CO2", "Carbon Dioxide"),
        ],
    )

    # First add_mod_class call
    gen_code.add_mod_class(
        module_cls="Test_Enum",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="gas_type",
                type="SignalRW",
                type_param="Test_EnumGas_typeEnum",
                path_annotation="balh:gas_type",
            )
        ],
        properties=[],
        cmd_plans=[],
        enum_classes=[enum1],
    )

    # Second add_mod_class call (same class name, different enum)
    gen_code.add_mod_class(
        module_cls="Test_Enum",
        bases=["SECoPDevice"],
        parameters=[
            ParameterAttribute(
                name="gas_type",
                type="SignalRW",
                type_param="Test_EnumGas_typeEnum",
                path_annotation="balh:gas_type",
            )
        ],
        properties=[],
        cmd_plans=[],
        enum_classes=[enum2],
    )

    # Should only have one module class, but with both enums
    assert len(gen_code.module_classes) == 1
    assert len(gen_code.module_classes[0].enums) == 2

    # Collect and merge enums
    merged = gen_code._collect_all_enums()

    # Should have one enum using SupersetEnum with all 5 unique members
    assert len(merged) == 1
    assert merged[0].base_enum_class == "SupersetEnum"
    assert len(merged[0].members) == 5  # AR, N2, H2, He, CO2

    member_names = {m.name for m in merged[0].members}
    assert member_names == {"AR", "N2", "H2", "HE", "CO2"}



import time

# mypy: disable-error-code="attr-defined"
import warnings

import bluesky.plan_stubs as bps
from bluesky.plans import count

from secop_ophyd.SECoPDevices import SECoPNodeDevice

warnings.filterwarnings(
    "ignore",
    message="""The method Broker.insert may be removed
    in a future release of databroker.""",
)


async def test_run_engine_count(RE, cryo_sim, cryo_node: SECoPNodeDevice):  # noqa: N803
    RE(count([cryo_node.cryo], num=5, delay=1))


async def test_run_engine_string_value(
    nested_struct_sim, RE, nested_node_no_re: SECoPNodeDevice  # noqa: N803
):
    def read_str():
        val = yield from bps.rd(nested_node_no_re.str_test.value)
        print(val)
        assert val is not None

    RE(read_str())


async def test_abs_set_wait_behaviour(
    RE, cryo_sim, cryo_node: SECoPNodeDevice  # noqa: N803
):

    def wait_on_abs_set():
        yield from bps.abs_set(cryo_node.cryo.window, 10, wait=True)
        yield from bps.abs_set(cryo_node.cryo.ramp, 20, wait=True)
        yield from bps.abs_set(cryo_node.cryo.tolerance, 2, wait=True)
        before = time.time()
        yield from bps.abs_set(cryo_node.cryo, 20, wait=True)
        after = time.time()

        assert after - before >= 10

        before = time.time()
        yield from bps.abs_set(cryo_node.cryo, 10, wait=False)
        after = time.time()

        assert after - before < 5

    RE(wait_on_abs_set())
